#!/usr/bin/env python3
r'''
------------------------------------------------------------------------------
    webquiz_makequiz | build and generate the quiz page using webquiz_xml
                        | and webquiz_layout
------------------------------------------------------------------------------
    Copyright (C) Andrew Mathas, University of Sydney

    Distributed under the terms of the GNU General Public License (GPL)
                  http://www.gnu.org/licenses/

    This file is part of the WebQuiz system.

    <Andrew.Mathas@sydney.edu.au>
    <Donald.Taylor@sydney.edu.au>
------------------------------------------------------------------------------
'''
import subprocess
import codecs
import shutil
import os
import re

import webquiz_templates
import webquiz_util
import webquiz_xml

#################################################################################
class MakeWebQuiz(object):
    """
    Convert a webquiz latex file to an online quiz.

    There are several steps:
      1. If given a LaTeX file then run htlatex/make4ht on the latex file to generate an
         xml file for the quiz that has all LaTeS markup converted to html.
      2. Read in the xml file version of the quiz
      3. Spit out the html version

    The HTMl is constructed using the template strings in webquiz_templates
    """
    # attributes that will form part of the generated web page
    header         = ''  # page header: title, meta data, links
    css            = ''  # css specifications
    javascript     = ''  # javascript code
    quiz_questions = ''  # the main quiz page
    side_menu      = ''  # the left hand quiz menu

    def __init__(self, quiz_name, quiz_file, options, settings, metadata):
        self.options = options
        self.settings = settings
        self.metadata = metadata
        self.quiz_file, extension = os.path.splitext(quiz_file)
        self.quiz_name = os.path.basename(self.quiz_file)
        self.webquiz_url = settings['webquiz_url']
        if self.webquiz_url[-1] == '/':
            self.webquiz_url =  self.webquiz_url[:len(self.webquiz_url)-1]

        # run htlatex only if quiz_file has a .tex extension
        if extension == '.tex':
            self.htlatex_quiz_file()

        self.read_xml_file()

        # use kpsewhich to fine the webquiz language file
        try:
            language_file = webquiz_util.kpsewhich('webquiz-{}.lang'.format(self.quiz.language))
        except subprocess.CalledProcessError:
            self.webquiz_error(
                'kpsewhich is unable to find language file for "{}"'.format(self.quiz.language)
            )
        # read the language file and store as a dictonary
        self.language = webquiz_util.MetaData(language_file)

        # initialise number of quiz and discussion items
        self.number_discussions = len(self.quiz.discussion_list)
        self.number_questions = len(self.quiz.question_list)

        # build the different components of the quiz web page
        self.add_meta_data()
        self.add_question_javascript()
        self.add_side_menu()
        self.add_quiz_header_and_questions()
        self.add_breadcrumbs()

        # add the initialisation warning if webquiz has not been initialised
        if self.settings.initialise_warning != '':
            self.breadcrumbs = self.settings.initialise_warning + self.breadcrumbs

        # now write the quiz to the html file
        with codecs.open(self.quiz_name + '.html', 'w', encoding='utf8') as file:
            # write the quiz in the specified format
            file.write(self.options.write_web_page(self))

    def webquiz_debug(self, msg):
        r'''
            Customised debugging message for the makequiz module
        '''
        webquiz_util.webquiz_debug(self.settings.debugging, 'makequiz: '+msg)

    def webquiz_error(self, msg, err=None):
        r'''
            Customised eror message for the makequiz module
        '''
        webquiz_util.webquiz_error(self.settings.debugging, 'makequiz: '+msg, err)

    def add_breadcrumbs(self):
        r'''
        Build the breadcrumbs by appropriately parsing the magic crumbs
        '''
        self.breadcrumbs = ''
        if self.quiz.breadcrumbs != '':
            # build the bread crumbs
            crumbs = '\n  '
            for crumb in self.quiz.breadcrumbs.split('|'):
                crumb = crumb.strip()

                if crumb == 'breadcrumb':
                    crumbs += self.add_breadcrumb_line(self.quiz.breadcrumb, missing='breadcrumb')

                elif crumb == 'department':
                    crumbs += self.add_breadcrumb_line(
                        text=self.quiz.department,
                        url=self.quiz.department_url,
                        missing='department')

                elif crumb == 'institution':
                    crumbs += self.add_breadcrumb_line(
                        text=self.quiz.institution,
                        url=self.quiz.institution_url,
                        missing='institution')

                elif crumb == 'quizindex':
                    if self.quiz.quiz_index == []:
                        crumbs += webquiz_templates.breadcrumb_quizindex.format(
                            quizzes_url=self.quiz.quizzes_url,
                            **self.language)
                    else:
                        crumbs += self.add_breadcrumb_line('Quizzes')

                elif crumb == 'Title':
                    crumbs += self.add_breadcrumb_line(text = self.quiz.title, missing='title')

                elif crumb == 'title':
                    title = self.quiz.title
                    crumbs += self.add_breadcrumb_line(
                                  text = title[:title.index(':')] if ':' in title else title,
                                  missing='title'
                              )

                elif crumb == 'unitcode':
                    crumbs += self.add_breadcrumb_line(
                        text=self.quiz.unit_code,
                        url=self.quiz.unit_url,
                        missing='unit code')

                elif crumb == 'unitname':
                    crumbs += self.add_breadcrumb_line(
                        text=self.quiz.unit_name,
                        url=self.quiz.unit_url,
                        missing='unit name')

                elif crumb != '':
                    lastSpace = crumb.rfind(' ')
                    url = crumb[lastSpace:].strip()
                    if url[0] == '/' or url.lower()[:4] == 'http':
                        crumbs += self.add_breadcrumb_line(text=crumb[:lastSpace], url=url)
                    else:
                        crumbs += self.add_breadcrumb_line(crumb)

            self.breadcrumbs = webquiz_templates.breadcrumbs.format(crumbs=crumbs)

    def add_breadcrumb_line(self, text, url='', missing='??'):
        r'''
        Return a line to add the bread crumbs, with errors if necessary
        '''
        if url == '':
            return webquiz_templates.breadcrumb_line_text.format(
                        text=text if text != '' else '?? ' + missing)

        return webquiz_templates.breadcrumb_line_url.format(
                    url=url, text=text if text != '' else '?? ' + missing)

    def htlatex_quiz_file(self):
        r'''
        Process the file using htlatex/make4ht. This converts the quiz to an xml
        with markup specifying the different elements of the quiz page.
        '''
        # at the minimum we put a css file into a <quiz_name> subdirectory
        if not os.path.exists(self.quiz_name):
            os.makedirs(self.quiz_name)

        try:
            self.options.talk('Processing {}.tex with TeX4ht'.format(self.quiz_name))
            # there is a slightly torturous process to convert the engine
            # settings into a command line option that make4ht understands
            cmd = 'make4ht --utf8 --config webquiz.cfg {draft} {engine} {escape} {make4ht_options} {quiz_file}.tex'.format(
                draft='--mode draft' if self.options.draft else '',
                engine=self.settings.settings['engine']['values'][self.options.engine],
                escape='--shell-escape' if self.options.shell_escape else '',
                make4ht_options=self.options.make4ht_options,
                quiz_file=self.quiz_file
            )
            self.options.run(cmd)

            # move the css file into the quiz_file subdirectory
            if os.path.exists(self.quiz_file + '.css'):
                shutil.move(
                    self.quiz_file + '.css',
                    os.path.join(self.quiz_name, self.quiz_name + '.css'))

            # Now move any images that were created into the quiz_name
            # subdirectory and update the links in the html file As htlatex
            # generates an html file, we rename this as an xml file at the same
            # time - in the cfg file, \Preamable{ext=xml} should lead to an xml
            # file being created but this doesn't seem to work ??
            try:
                fix_img = re.compile(r'^(|.* )\b(data|src)="([-0-9a-zA-Z]*\.(?:png|svg))" (.*)$')
                with codecs.open(self.quiz_file + '.html', 'r', encoding='utf8') as make4ht_file:
                    with codecs.open(self.quiz_name + '.xml', 'w', encoding='utf8') as xml_file:
                        for line in make4ht_file:
                            match = fix_img.match(line)
                            if match is None:
                                xml_file.write(line)
                            else:
                                # update html link and move file
                                start, src, image, rest_of_line = match.groups()
                                xml_file.write(r'{}{}="{}/{}" {}'.format(
                                    start, src, self.quiz_name, image, rest_of_line))
                                shutil.move(image, os.path.join(self.quiz_name, image))

            except OSError as err:
                self.webquiz_error(
                    'there was a problem moving the image files for {}'.format(
                        self.quiz_name), err)

        except Exception as err:
            self.webquiz_error( 'something went wrong when running htlatex on {}'.format(self.quiz_file), err)

    def read_xml_file(self):
        r'''
        Read in the webquiz xml file for the quiz and store the xml document
        tree in ``self.quiz``.
        '''
        try:
            # read in the xml version of the quiz
            if not os.path.isfile(self.quiz_name + '.xml'):
                self.webquiz_error('{}.xml does not exist!?'.format(self.quiz_name))
            self.quiz = webquiz_xml.ReadWebQuizXmlFile(self.quiz_name + '.xml', self.settings)
        except Exception as err:
            self.webquiz_error('error reading the xml generated for {}. Please check your latex source.'
                .format(self.quiz_name), err)

    def add_meta_data(self):
        """ add the meta data for the web page to self.header """
        # meta tags`
        self.header += webquiz_templates.html_meta.format(
            version=self.metadata.version,
            authors=self.metadata.authors,
            webquiz_url=self.webquiz_url,
            description=self.metadata.description,
            copyright=self.metadata.copyright,
            department=self.quiz.department,
            institution=self.quiz.institution,
            quiz_file=self.quiz_name,
            theme=self.quiz.theme)
        if self.quiz.mathjs:
            self.header += webquiz_templates.mathjs

    def add_side_menu(self):
        """ construct the left hand quiz menu """
        if self.quiz.discussion_list != []:  # links for discussion items
            discussion_list = '\n       <ul>\n   {}\n       </ul>'.format(
                '\n   '.join(
                    webquiz_templates.discuss.format(b=q + 1, title=d.short_heading)
                    for (q, d) in enumerate(self.quiz.discussion_list)))
        else:
            discussion_list = ''

        # department and institution links
        department = '''<a href="{0.department_url}">{0.department}</a>'''.format(self.quiz)
        institution = '''<a href="{0.institution_url}">{0.institution}</a>'''.format(self.quiz)

        # question buttons
        if self.number_questions == 0:
            question_buttons = ''
        else:
            buttons = '\n' + '\n'.join(
                webquiz_templates.button.format(b=q,
                    cls='button-selected' if self.quiz.discussion_list==[] and q==1 else 'blank'
                )
                for q in range(1, self.number_questions + 1))
            question_buttons=webquiz_templates.question_buttons.format(
                buttons=buttons, **self.language
            )

        # the full side menu
        self.side_menu = webquiz_templates.side_menu.format(
            discussion_list=discussion_list,
            version=self.metadata.version,
            department=department,
            institution=institution,
            side_questions=self.language['questions'] if self.number_questions>0 else '',
            question_buttons=question_buttons,
            copyright_years=self.metadata.copyright[:self.metadata.copyright.index(' ')],
            **self.language)

    def add_question_javascript(self):
        """
        Add the javascript for the questions to self and write the javascript
        initialisation file, <quiz>/quiz_specs.js, for the quiz.  When the quiz
        page is loaded, WebQuizInit reads the quiz_specs initialisation file to
        load the answers to the questions,  and the headers for the discussion
        items. We don't explicitly list quiz_specs.js in the meta data for the
        quiz page because we want to hide this information from the student,
        although they can easily get this if they open by the javascript
        console and know what to look for.
        """

        try:
            os.makedirs(self.quiz_name, exist_ok=True)
            os.chmod(self.quiz_name, mode=0o755)
            with codecs.open(os.path.join(self.quiz_name, 'wq-' + self.quiz_name + '.js'), 'w',
                             encoding='utf8') as quiz_specs:
                if self.number_discussions > 0:
                    for (i, d) in enumerate(self.quiz.discussion_list):
                        quiz_specs.write('Discussion[{}]="{}";\n'.format(i+1, d.heading))
                if self.number_questions > 0:
                    for (i, question) in enumerate(self.quiz.question_list):
                        # QuizSpecifications is a 0-based array
                        quiz_specs.write('QuizSpecifications[{}]=[];\n'.format(i+1))
                        quiz_specs.write('QuizSpecifications[{}].type="{}";\n'.format(i+1, question.type))
                        if question.type == 'input':
                            quiz_specs.write('QuizSpecifications[{}].value="{}";\n'.format(i+1,
                                question.answer.lower() if question.comparison=='lowercase'
                                                        else question.answer
                              )
                            )
                            quiz_specs.write('QuizSpecifications[{}].comparison="{}";\n'.format(i+1, question.comparison))
                        else:
                            quiz_specs.write(''.join(
                                    'QuizSpecifications[{}][{}]={};\n'.format(i+1, j, s.correct)
                                    for (j, s) in enumerate(question.items)
                                )
                            )

                if self.quiz.hide_side_menu:
                    quiz_specs.write('toggle_side_menu();\n')
                if self.quiz.one_page:
                    quiz_specs.write('onePage = true;\n')
                if self.quiz.random_order:
                    quiz_specs.write('shuffleQuestions();\n')
                quiz_specs.write('initSession();\n')
                if self.number_discussions+self.number_questions>0:
                    quiz_specs.write('gotoQuestion({});'.format(
                        -1 if self.number_discussions>0 else 1)
                    )

        except Exception as err:
            self.webquiz_error('error writing quiz specifications', err)

        self.javascript = webquiz_templates.questions_javascript.format(
            webquiz_url=self.webquiz_url,
            mathjax=self.settings['mathjax']
        )
        self.webquiz_init = webquiz_templates.webquiz_init.format(
            number_questions=self.number_questions,
            number_discussions=self.number_discussions,
            quiz_file=self.quiz_name,
        )

    def add_quiz_header_and_questions(self):
        r'''
        Write the quiz head and the main body of the quiz.
        '''
        if self.quiz.one_page:
            arrows = ''
        else:
            arrows = webquiz_templates.navigation_arrows.format(
                        question_number=self.quiz.discussion_list[0].heading
                                    if self.quiz.discussion_list != []
                                    else '1' if self.quiz.question_list > []
                                    else '',
                        **self.language
                    )

        # specify the quiz header - this will be wrapped in <div class="question-header>...</div>
        self.quiz_header = webquiz_templates.quiz_header.format(
            title=self.quiz.title,
            arrows=arrows,
            **self.language
        )

        # index for quiz
        if self.quiz.quiz_index != []:
            # add index to the web page
            self.quiz_questions += webquiz_templates.quiz_index_div.format(
                title=self.quiz.title if self.quiz.title!='' else self.quiz.unit_name,
                quiz_index='\n          '.join(
                    webquiz_templates.index_item.format(
                        url=q.url,
                        title='{} {}. {}'.format(self.language['quiz'],num+1,q.title)
                                if q.prompt else q.title,
                    ) for (num, q) in enumerate(self.quiz.quiz_index)),
                **self.language)
            # write a javascript file for displaying the menu
            # quizmenu = the index file for the quizzes in this directory
            with codecs.open('quizindex.js', 'w', encoding='utf8') as quizmenu:
                quizmenu.write('var QuizTitles = [\n{titles}\n];\n'.format(
                    titles=',\n'.join("  ['{}', '{}']".format(
                             '{} {}. {}'.format(self.language['quiz'],num+1,q.title) 
                                       if q.prompt else q.title,
                              q.url)
                        for (num,q) in enumerate(self.quiz.quiz_index))
                    )
                )
                quizmenu.write(webquiz_templates.create_quizindex_menu)

        # now comes the main page text
        # discussion(s) masquerade as negative questions
        if self.quiz.discussion_list != []:
            dnum = 0
            for d in self.quiz.discussion_list:
                dnum += 1
                self.quiz_questions += webquiz_templates.discussion.format(
                    dnum=dnum,
                    discussion=d,
                    display='inline' if self.quiz.one_page else 'none',
                    heading=webquiz_templates.discussion_heading.format(d.heading)
                            if self.quiz.one_page else ''
                )

        # finally we print the questions
        if self.quiz.question_list != []:
            self.quiz_questions += ''.join(
                webquiz_templates.question_wrapper.format(
                    qnum=qnum + 1,
                    question_number='{} {}. '.format(self.language.question, qnum+1)
                                    if self.quiz.one_page else '',
                    display='inline' if self.quiz.one_page else 'none',
                    question=self.print_question(quiz_question, qnum + 1),
                    feedback=self.print_feedback(quiz_question, qnum + 1)
                )
                for (qnum, quiz_question) in enumerate(self.quiz.question_list))

    def print_question(self, question, qnum):
        r'''Here:
            - question is the question
            - qnum is the number of the question
        '''
        if question.type == 'input':
            self.webquiz_debug('Q{}: after_text={}.'.format(qnum, question.after_text))
            question_options = webquiz_templates.input_answer.format(
                                 size=5+len('{}'.format(question.answer)),
                                 after_text=question.after_text,
                                 qnum=qnum,
                                 answer=self.language['answer']+':' if question.prompt else ''
            )
        elif question.type in ['single', 'multiple']:
            question_options = webquiz_templates.choice_answer.format(
                    after_text=question.after_text,
                    choices='\n'.join(self.print_choices(qnum, question, choice)
                        for choice in range(len(question.items)))
            )
        else:
            self.webquiz_error('Unknown question type "{}" in question {}'.format(question.type, qnum))
        return webquiz_templates.question_text.format(
            qnum=qnum,
            question_text=question.text,
            nextquestion='' if self.quiz.one_page else webquiz_templates.nextquestion.format(**self.language),
            question_options=question_options,
            **self.language)

    def print_choices(self, qnum, question, part):
        r'''
        Here:
            - qnum     = question number
            - question = current question
            - part     = number of the option we need to process.
        We put the parts into question.columns multicolumn format, so we have
        to add '<tr>' and '</tr>' tags depending on part.
        '''
        choice = question.items[part]
        item = '<tr>' if question.columns == 1 or (part % question.columns) == 0 else '<td>&nbsp;</td>'
        if question.type == 'single':
            item += webquiz_templates.single_item.format(choice=choice.symbol, qnum=qnum, text=choice.text)
        elif question.type == 'multiple':
            item += webquiz_templates.multiple_item.format(
                choice=choice.symbol,
                qnum=qnum,
                optnum=part,
                text=choice.text
            )
        else:
            item += '<!-- internal error: %s -->\n' % question.type
            self.webquiz_error('Unknown question type "{}" in question {}'.format(question.type, qnum))

        if question.columns == 1 or (part+1) % question.columns == 0 or part == len(question.items) - 1:
            item += '   </tr><!-- part={}, cols={}, # answers = {} -->\n'.format(
                part, question.columns, len(question.items))
        return item

    def print_feedback(self, question, qnum):
        r'''
        Generate the HTML for displaying the feedback help text when the user
        answers a question.
        '''
        if question.type == 'input':
            feedback = webquiz_templates.tf_feedback_text.format(
                choice=qnum,
                feedback='true',
                correct_answer=self.language.correct,
                answer2='',
                text=question.feedback_right)
            feedback += webquiz_templates.tf_feedback_text.format(
                choice=qnum,
                feedback='false',
                correct_answer=self.language.incorrect,
                answer2=self.language.try_again,
                text=question.feedback_wrong)
        elif question.type == "single":
            feedback = '\n' + '\n'.join(
                webquiz_templates.single_feedback.format(
                    qnum=qnum,
                    part=snum + 1,
                    correct_answer=self.language.correct if s.correct == 'true' else self.language.incorrect,
                    alpha_choice=self.language.choice.format(s.symbol),
                    feedback=s.feedback,
                    **self.language)
                for (snum, s) in enumerate(question.items))
        elif question.type == "multiple":
            feedback = '\n' + '\n'.join(webquiz_templates.multiple_feedback.format(
                qnum=qnum,
                part=snum + 1,
                correct_answer=getattr(self.language, s.correct).capitalize(),
                feedback=s.feedback,
                multiple_choice_opener=self.language.multiple_incorrect.
                format(s.symbol),
                **self.language)
                for (snum, s) in enumerate(question.items)
            )
            feedback += webquiz_templates.multiple_feedback_correct.format(
                qnum=qnum,
                feedback='\n'.join(webquiz_templates.multiple_feedback_answer.format(
                                        correct_answer=getattr(self.language, s.correct).capitalize(),
                                        reason=s.feedback) for s in question.items),
                                    **self.language)
        else:
            self.webquiz_error('Unknown question type "{}" in question {}'.format(question.type, qnum))

        return '<div class="answer">' + feedback + '</div>'



#!/usr/bin/env python3
r'''
------------------------------------------------------------------------------
    webquiz_util | utilty functions 
------------------------------------------------------------------------------
    Copyright (C) Andrew Mathas, University of Sydney

    Distributed under the terms of the GNU General Public License (GPL)
                  http://www.gnu.org/licenses/

    This file is part of the WebQuiz system.

    <Andrew.Mathas@sydney.edu.au>
------------------------------------------------------------------------------
'''

import os
import subprocess
import shutil
import stat
import sys
import traceback

# ---------------------------------------------------------------------------------------
# Return the full path for a file in the webquiz directory
webquiz_file = lambda file: os.path.join(os.path.dirname(os.path.realpath(__file__)), file)

def shell_command(cmd):
    r'''
    Short-cut for shell commands
    '''
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True).decode('ascii').strip()

def kpsewhich(search):
    r'''
    Short-cut to access kpsewhich output.

    usage: kpsewhich('-var-value=TEXMFLOCAL')
    '''
    return subprocess.check_output('kpsewhich ' + search, shell=True).decode('ascii').strip()

# ---------------------------------------------------------------------------------------
class MetaData(dict):
    r"""
    A dummy class for reading, accessing and storing key-value pairs from
    a file. Any internal spaces in the key name are replaced with underscores
    and lines without a key-value pair are ignored.

    The key-value pairs are available as both attributes and items

    Usage: MetaData(filename) or MetaData(filename, key-value pairs)
    """

    def __init__(self, filename, **args):
        dict.__init__(self)
        # add attributes from **args
        for key in args:
            setattr(self, key, args[key])
        with open(filename, 'r') as meta:
            for line in meta:
                if '=' in line:
                    key, val = line.strip().split('=')
                    if key.strip() != '':
                        self.__setitem__(key.strip().lower().replace(' ', '_'),
                                         val.strip())
                        setattr(self,
                                key.strip().lower().replace(' ', '_'),
                                val.strip())


#################################################################################
def webquiz_debug(debugging, *arg):
    if debugging:
        sys.stderr.write(' '.join('{}'.format(a) for a in arg)+'\n')


#################################################################################
def webquiz_error(debugging, msg, err=None):
    r'''
    Consistent handling of errors in magthquiz: print the message `msg` and
    exist with error code `err.errno` if it is available.abs
    '''
    print('{dash}WebQuiz error:\n  {msg}\n{dash}'.format(
           msg=msg, dash='-'*40+'\n')
    )

    if err is not None:
        trace = traceback.extract_tb(sys.exc_info()[2])
        filename, lineno, fn, text = trace[-1]
        print('File: {}, line number: {}\nError {} in {}: {}'.format(
            filename, lineno, err, fn, text))

    if debugging and err is not None:
        raise

    if hasattr(err, 'errno'):
        sys.exit(err.errno)

    sys.exit(1)


###############################################################################
def copytree(src, dst, symlinks=False, ignore=None):
    r''' Recursively copy directory tree, fixing shutil.copytree
         from https://stackoverflow.com/questions/1868714
    '''
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    if ignore:
        excl = ignore(src, lst)
        lst = [x for x in lst if x not in excl]
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if symlinks and os.path.islink(s):
            if os.path.lexists(d):
                os.remove(d)
            os.symlink(os.readlink(s), d)
            try:
                st = os.lstat(s)
                mode = stat.S_IMODE(st.st_mode)
                os.lchmod(d, mode)
            except OSError:
                pass  # lchmod not available
        elif os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)

###############################################################################
## shortcuts for running commands
environ = os.environ.copy()

def run(cmd, shell=False):
    r'''
    Run commands with output to stdout and errors to stderr
    '''
    if shell:
        subprocess.call(cmd, env=environ, shell=True)
    else:
        subprocess.call(cmd.split(), env=environ)

def quiet_run(cmd, shell=False):
    r'''
    Run commands with ignoring and sending errors to stderr
    '''
    if shell:
        subprocess.call(cmd, env=environ, stdout=open(os.devnull, 'wb'), shell=True)
    else:
        subprocess.call(cmd.split(), env=environ, stdout=open(os.devnull, 'wb'))

def silent_run(cmd, shell=False):
    r'''
    Run commands ignoring all output and errors
    '''
    if shell:
        subprocess.call(cmd, env=environ, stdout=open(os.devnull, 'wb'), stderr=open(os.devnull, 'wb'), shell=True)
    else:
        subprocess.call(cmd.split(), env=environ, stdout=open(os.devnull, 'wb'), stderr=open(os.devnull, 'wb'))



#!/usr/bin/env python3
r'''
------------------------------------------------------------------------------
    webquiz | Online quizzes generated from LaTeX using python and TeX4ht
               | This module mainly deals with command-line options and
               | settings and then calls MakeWebQuiz to build the quiz
------------------------------------------------------------------------------
    Copyright (C) Andrew Mathas and Donald Taylor, University of Sydney

    Distributed under the terms of the GNU General Public License (GPL)
                  http://www.gnu.org/licenses/

    This file is part of the WebQuiz system.

    <Andrew.Mathas@sydney.edu.au>
------------------------------------------------------------------------------
'''

import argparse
import codecs
import errno
import glob
import os
import re
import shutil
import signal
import subprocess
import sys

# imports of webquiz code
import webquiz_makequiz
import webquiz_templates
import webquiz_util

#################################################################################
# read in basic meta data such as author, version, ... and set debugging=False
try:
    metadata = webquiz_util.MetaData(webquiz_util.kpsewhich('webquiz.ini'), debugging=False)
except subprocess.CalledProcessError:
    # check to see if we are running from the zip file
    ini_file = os.path.join(webquiz_util.webquiz_file(''), '..', 'latex', 'webquiz.ini')
    try:
        metadata = webquiz_util.MetaData(ini_file, debugging=False)
    except (FileNotFoundError, subprocess.CalledProcessError):
        print(webquiz_templates.missing_webquiz_ini.format(ini_file))
        sys.exit(1)

# ---------------------------------------------------------------------------------------
def graceful_exit(sig, frame):
    ''' exit gracefully on SIGINT and SIGTERM'''
    if metadata:
        webquiz_util.webquiz_error(False, 'program terminated (signal {}\n  {})'.format(sig, frame))
    else:
        webquiz_util.webquiz_error(False, 'program terminated (signal {})'.format(sig))

signal.signal(signal.SIGINT, graceful_exit)
signal.signal(signal.SIGTERM, graceful_exit)


#################################################################################
def preprocess_with_pst2pdf(options, quiz_file):
    r'''
    Preprocess the latex file using pst2pdf. As we are preprocessing the file it
    is not enough to have latex pass us a flag that tells us to use pst2pdf.
    Instead, we have to extract the class file option from the tex file

    INPUT: quiz_file should be the name of the quiz file, WITHOUT the .tex extension
    '''
    options.talk('Preprocessing {} with pst2pdsf'.format(quiz_file))
    try:
        # pst2pdf converts pspicture environments to svg images and makes a
        # new latex file quiz_file+'-pdf' that includes these
        cmd = 'pst2pdf --svg --imgdir={q_file} {q_file}.tex'.format(q_file=quiz_file)
        # pst2pdf is missing a #!-header so we need shell=True
        options.run(cmd, shell=True)
    except OSError as err:
        if err.errno == errno.ENOENT:
            webquiz_util.webquiz_error(options.debugging, 'pst2pdf not found. You need to install pst2pdf to use the pst2pdf option', err)
        else:
            webquiz_util.webquiz_error(options.debugging, 'error running pst2pdf on {}'.format(quiz_file), err)

    # match \includegraphics commands
    fix_svg = re.compile(r'(\\includegraphics\[scale=1\])\{('+quiz_file+r'-fig-[0-9]*)\}')
    # the svg images are in the quiz_file subdirectory but latex can't
    # find them so we update the tex file to look in the right place
    try:
        with codecs.open(quiz_file + '-pdf.tex', 'r', encoding='utf8') as pst_file:
            with codecs.open(quiz_file + '-pdf-fixed.tex', 'w', encoding='utf8') as pst_fixed:
                for line in pst_file:
                    pst_fixed.write(fix_svg.sub(r'\1{%s/\2.svg}' % quiz_file, line))
    except OSError as err:
        webquiz_util.webquiz_error(options.debugging,
            'there was an problem running pst2pdf for {}'.format(quiz_file),
            err
        )

class WebQuizSettings:
    r'''
    Class for initialising webquiz. This covers both reading and writing the
    webquizrc file and copying files into the web directories during
    initialisation. The settings themselves are stored in the attribute
    settings, which is a dictionary. The class reads and writes the settings to
    the webquizrc file and the values of the settings are available as items:
        >>> wq = WebQuizSettings()
        >>> wq['webquiz_url']
        ... /WebQuiz
        >>> wq['webquiz_url'] = '/new_url'
    '''

    # default of settings for the webquizrc file - a dictionary of dictionaries
    # the 'help' field is for printing descriptions of the settings to help the
    # user - they are also printed in the webquizrc file
    settings = dict(
        webquiz_url={
            'default': '',
            'advanced': False,
            'help': 'Relative URL for the webquiz web directory',
        },
        webquiz_www={
            'default': '',
            'advanced': False,
            'help': 'Full path to WebQuiz web directory',
        },
        language={
            'default': 'english',
            'advanced': False,
            'help': 'Default language used on web pages'
        },
        engine = {
            'default': 'latex',
            'advanced': False,
            'help': 'Default TeX engine used to compile web pages',
            'values': dict(latex='', lua='--lua', xelatex='--xetex')
        },
        theme={
            'default': 'default',
            'advanced': False,
            'help': 'Default colour theme used on web pages'
        },
        breadcrumbs={
            'default': '',
            'advanced': False,
            'help': 'Breadcrumbs at the top of quiz page',
        },
        department={
            'default': '',
            'advanced': False,
            'help': 'Name of department',
        },
        department_url={
            'default': '/',
            'advanced': False,
            'help': 'URL for department',
        },
        institution={
            'default': '',
            'advanced': False,
            'help': 'Institution or university',
        },
        institution_url={
            'default': '/',
            'advanced': False,
            'help': 'URL for institution or university',
        },
        hide_side_menu={
            'default': 'false',
            'advanced': False,
            'help': 'Do not display the side menu at start of quiz',
        },
        one_page={
            'default': 'false',
            'advanced': False,
            'help': 'Display questions on one page',
        },
        random_order={
            'default': 'false',
            'advanced': False,
            'help': 'Randomly order the quiz questions',
        },
        webquiz_layout={
            'default': 'webquiz_layout',
            'advanced': True,
            'help': 'Name of python module that formats the quizzes',
        },
        make4ht={
            'default': '',
            'advanced': True,
            'help': 'Build file for make4ht',
        },
        mathjax={
            'default':
            'https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.1/MathJax.js',
            'advanced':
            True,
            'help':
            'URL for mathjax',
        },
        version={
            'advanced': False,
            'help': 'WebQuiz version number for webquizrc settings',
        })

    # by default we assume we don't need to print a initialisation warning
    initialise_warning = ''

    # turn debugging on by default because any error message that we hit before
    # we process the command line options really should not happen
    debugging = True

    # keep track of whether we have initialised
    have_initialised = False

    def __init__(self):
        '''
        First read the system webquizrc file and then read the
        to use some system settings and to override others.

        By default, there is no webquiz initialisation file. We first
        look for webquizrc in the webquiz source directory and then
        for .webquizrc file in the users home directory.
        '''
        self.settings['version']['default'] = metadata.version
        for key in self.settings:
            self.settings[key]['value'] = self.settings[key]['default']
            if not 'editable' in self.settings[key]:
                self.settings[key]['editable'] = False

        # define user and system rc file and load the ones that exist

        TEXMFLOCAL=''
        try:
            TEXMFLOCAL = webquiz_util.kpsewhich('-var-value TEXMFLOCAL')
        except subprocess.CalledProcessError:
            pass

        if TEXMFLOCAL == '':
            TEXMFLOCAL = webquiz_util.kpsewhich('-var-value TEXMFMAIN')

        self.system_rcfile = os.path.join(TEXMFLOCAL, 'tex', 'latex', 'webquiz', 'webquizrc')
        self.read_webquizrc(self.system_rcfile)

        # the user rc file defaults to:
        #   ~/.dotfiles/config/webquizrc if .dotfiles/config exists
        #   ~/.config/webquizrc if .config exists
        # and otherwise to ~/.webquizrc
        if os.path.isdir(os.path.join(os.path.expanduser('~'), '.dotfiles', 'config')):
            self.user_rcfile = os.path.join(os.path.expanduser('~'), '.dotfiles', 'config', 'webquizrc')
        elif os.path.isdir(os.path.join(os.path.expanduser('~'), '.config')):
            self.user_rcfile = os.path.join(os.path.expanduser('~'), '.config', 'webquizrc')
        else:
            self.user_rcfile = os.path.join(os.path.expanduser('~'), '.webquizrc')

        self.read_webquizrc(self.user_rcfile)

    def webquiz_debug(self, msg):
        r'''
            Customised debugging message for the MakeSettings module
        '''
        webquiz_util.webquiz_debug(self.debugging, 'main: '+msg)

    def webquiz_error(self, msg, err=None):
        r'''
            Customised error messages for the Module
        '''
        webquiz_util.webquiz_error(self.debugging, 'settings: '+msg, err)

    def __getitem__(self, key):
        r'''
        Return the value of the corresponding setting. That is, it returns
            self.settings[key]['value']
        and an error if the key is unknown.
        '''
        if key in self.settings:
            return self.settings[key]['value']

        self.webquiz_error('getitem: unknown setting "{}" in webquizrc.'.format(key))

    def __setitem__(self, key, value):
        r'''
        Set the value of the corresponding setting. This is the equivalent of
            self.settings[key]['value'] = value
        and an error if the key is unknown.
        '''
        if key in self.settings:
            self.settings[key]['value'] = value
        else:
            self.webquiz_error('setitem: unknown setting "{}" in webquizrc'.format(key))

    def read_webquizrc(self, rcfile, must_exist=False):
        r'''
        Read the settings from the specified webquizrc file - if it exists, in
        which case set self.rcfile equal to this directory. If the file does
        not exist then return without changing the current settings.
        '''
        if os.path.isfile(rcfile):
            try:
                with codecs.open(rcfile, 'r', encoding='utf8') as webquizrc:
                    for line in webquizrc:
                        if '#' in line:  # remove comments
                            line = line[:line.index('#')]
                        if '=' in line:
                            key, value = line.split('=')
                            key = key.strip().lower().replace('-','_')
                            value = value.strip()
                            if key in self.settings:
                                if value != self[key]:
                                    self[key] = value
                            elif key != '':
                                self.webquiz_error('unknown setting "{}" in {}'.format(key, rcfile))

                # record the rcfile for later use
                self.rcfile = rcfile

            except OSError as err:
                self.webquiz_error('there was a problem reading the rc-file {}'.format(rcfile), err)

            except Exception as err:
                self.webquiz_error('there was an error reading the webquizrc file,', err)

        elif must_exist:
            # this is only an error if we have been asked to read this file
            self.webquiz_error('the rc-file "{}" does not exist'.format(rcfile))

    def keys(self):
        r'''
        Return a list of keys for all settings, ordered alphabetically with the
        advanced options last/
        '''
        return sorted(self.settings.keys(), key=lambda k: '{}{}'.format(self.settings[k]['advanced'], k))

    def write_webquizrc(self):
        r'''
        Write the settings to the webquizrc file, defaulting to the user
        rcfile if unable to write to the system rcfile
        '''
        if not hasattr(self, 'rcfile'):
            # when initialising an rcfile will not exist yet
            self.rcfile = self.system_rcfile

        file_not_written = True
        while file_not_written:
            try:
                dire = os.path.dirname(self.rcfile)
                if dire != '' and not os.path.isdir(dire):
                    os.makedirs(dire, exist_ok=True)
                with codecs.open(self.rcfile, 'w', encoding='utf8') as rcfile:
                    for key in self.keys():
                        # Only save settings in the rcfile if they have changed
                        # Note that changed means changed from the last read
                        # rcfile rather than from the default (of course, the
                        # defaults serve as the "initial rcfile")
                        if key == 'version' or self.settings[key]['default']!=self[key]:
                            rcfile.write('# {}\n{:<17} = {}\n'.format(
                                           self.settings[key]['help'],
                                           key.replace('_','-'),
                                           self[key])
                            )

                print('\nWebQuiz settings saved in {}\n'.format( self.rcfile))
                input('Press RETURN to continue... ')
                file_not_written = False

            except (OSError, PermissionError) as err:
                # if writing to the system_rcfile then try to write to user_rcfile
                alt_rcfile = self.user_rcfile if self.rcfile != self.user_rcfile else self.system_rcfile
                response = input(
                    webquiz_templates.rc_permission_error.format(
                        error=err,
                        rcfile=self.rcfile,
                        alt_rcfile=alt_rcfile))
                if response.startswith('2'):
                    self.rcfile = alt_rcfile
                elif response.startswith('3'):
                    rcfile = input('WebQuiz rc-file: ')
                    print('\nTo access this rc-file you will need to use: webquiz --rcfile {} ...'.format(rcfile))
                    self.rcfile = os.path.expanduser(rcfile)
                elif not response.startswith('1'):
                    print('exiting...')
                    sys.exit(1)

    def list_settings(self, setting='all'):
        r'''
        Print the non-default settings for webquiz from the webquizrc
        '''
        if not hasattr(self, 'rcfile'):
            print(webquiz_templates.initialise_settings)
            sys.exit(1)

        if setting not in ['all', 'verbose', 'help']:
            setting = setting.replace('-', '_')
            if setting in self.settings:
                print(self.settings[setting]['value'])
            else:
                self.webquiz_error('{} is an invalid setting'.format(setting))

        elif setting=='all':
            dash = '-'*len('WebQuiz rc-file: {}'.format(self.rcfile))
            print('{dash}\nWebQuiz rc-file: {rcfile}\n{dash}'.format(rcfile=self.rcfile, dash=dash))
            for key in self.keys():
                print('{:<17} = {}'.format(key.replace('_', '-'), self[key]))
            print('{dash}'.format(dash=dash))

        elif setting=='help':
            for key in self.keys():
                print('{:<17} {}'.format(key.replace('_', '-'), self.settings[key]['help'].lower()))

        else:
            print('WebQuiz settings from {}'.format(self.rcfile))
            for key in self.keys():
                print('# {}{}\n{:<17} = {:<17}  {}'.format(
                        self.settings[key]['help'],
                        ' (advanced)' if self.settings[key]['advanced'] else '',
                        key.replace('_', '-'),
                        self[key],
                        '(default)' if self[key]==self.settings[key]['default'] else ''
                        )
                )

    def initialise_webquiz(self, need_to_initialise=False, developer=False):
        r'''
        Set the root for the WebQuiz web directory and copy the www files into
        this directory. Once this is done save the settings to webquizrc.
        This method should only be used when WebQuiz is being set up.

        If `need_to_initialise` is `True` then this is a forced initialisation.
        '''

        # keep track of whether we have initialised
        self.have_initialised = True

        if need_to_initialise:
            self.initialise_warning = webquiz_templates.web_initialise_warning
            initialise = input(webquiz_templates.initialise_invite)
            if initialise!='' and initialise.strip().lower()[0]!='y':
                self['webquiz_url'] = 'http://www.maths.usyd.edu.au/u/mathas/WebQuiz'
                return

        if self['webquiz_url']=='':
            self['webquiz_url'] = '/WebQuiz'

        # prompt for directory and copy files - are these reasonable defaults
        # for each OS?
        if sys.platform == 'darwin':
            default_root = '/Library/WebServer/Documents/WebQuiz'
            platform = 'Mac OSX'
        elif sys.platform.startswith('win'):
            default_root = 'c:\inetpub\wwwroot\WebQuiz'
            platform = 'Windows'
        else:
            default_root = '/var/www/html/WebQuiz'
            platform = sys.platform.capitalize()

        if self['webquiz_www'] != '':
            webquiz_root = self['webquiz_www']
        else:
            webquiz_root = default_root

        print(webquiz_templates.initialise_introduction)
        input('Press RETURN to continue... ')

        print(webquiz_templates.webroot_request.format(
                platform=platform,
                webquiz_dir = webquiz_root)
        )
        input('Press RETURN to continue... ')

        files_copied = False
        while not files_copied:
            web_dir = input('\nWebQuiz web directory:\n[{}] '.format(webquiz_root))
            if web_dir == '':
                web_dir = webquiz_root
            else:
                web_dir = os.path.expanduser(web_dir)

            print('Web directory set to {}'.format(web_dir))
            if web_dir=='SMS':
                # undocumented: allow links to SMS web pages
                self['webquiz_www'] = 'SMS'
                self['webquiz_url'] = 'http://www.maths.usyd.edu.au/u/mathas/WebQuiz'

            else:
                try:
                    # ...remove the doc directory
                    web_doc = os.path.join(web_dir, 'doc')
                    if os.path.isfile(web_doc) or os.path.islink(web_doc):
                        os.remove(web_doc)
                    elif os.path.isdir(web_doc):
                        shutil.rmtree(web_doc)

                    # Need to locate the www directory, which should be a subdirectory
                    # of the webquiz doc directory. First try using texdoc
                    webquiz_doc = ''
                    try:
                        webquiz_pdf = webquiz_util.shell_command('texdoc --list --machine webquiz.pdf').split()[-1]
                        if webquiz_pdf.endswith('webquiz.pdf'):
                            webquiz_doc = os.path.dirname(webquiz_pdf)
                    except subprocess.CalledProcessError:
                        pass

                    # if texdoc failed then try using TEXMFMAIN
                    if webquiz_doc=='':
                        try:
                            webquiz_doc = os.path.join(webquiz_util.kpsewhich('-var-value TEXMFMAIN'), 'doc','latex', 'webquiz')
                        except subprocess.CalledProcessError:
                            pass

                    # if we still don't have webquiz_doc then try working backwards from webquiz.cls
                    # unlikely to work if TEXMFMAIN doesn't
                    if not os.path.isdir(webquiz_doc):
                        parent = os.path.dirname
                        try:
                            texdist_dir = parent(parent(parent(parent(parent(webquiz_util.kpsewhich('webquiz.cls'))))))
                        except subprocess.CalledProcessError:
                            print(webquiz_templates.not_installed.format(metadata.repository))
                            sys.exit(1)

                        webquiz_doc = os.path.join(texdist_dir, 'doc', 'latex', 'webquiz')

                    # get the root directory of the source code for developer
                    # mode and just in case webquiz_www still does not exist
                    webquiz_src = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

                    webquiz_www = os.path.join(webquiz_doc, 'www')
                    if not os.path.isdir(webquiz_www):
                        webquiz_www = os.path.join(webquiz_src, 'doc', 'www')

                    if developer and os.path.isdir(os.path.join(webquiz_src, 'doc')):
                        # this is a development version so add links from the
                        # web directory to the css,doc and js directories
                        print('\nInstalling files for development version')
                        print('Linking web files {} -> {} ...\n'.format(web_dir, webquiz_src))
                        if not os.path.exists(web_dir):
                            os.makedirs(web_dir)

                        for (src, target) in [('javascript', 'js'), ('css', 'css'), ('doc', 'doc')]:
                            newlink = os.path.join(web_dir, target)
                            try:
                                os.remove(newlink)
                            except FileNotFoundError:
                                pass
                            try:
                                os.symlink(os.path.join(webquiz_src,src), newlink)
                            except OSError as err:
                                print('There was a problem linking {}: {}'.format(newlink, err))

                    else:
                        # loop until we find some files to install or exit
                        while not os.path.isdir(webquiz_www) or webquiz_www=='':
                            print('\nUnable to find the WebQuiz web files')
                            webquiz_www = input('Please enter the location of the WebQuiz www directory\nor press RETURN to exit: ')
                            webquiz_www = os.path.expanduser(webquiz_www)
                            if webquiz_www=='':
                                sys.exit()
                            if not (webquiz_www.endswith('www/') or webquiz_www.endswith('www')):
                                print('\nThe webquiz web directory is called www, so\n  {}\ncannot be the right directory.'.format(
                                      webquiz_www)
                                )
                                webquiz_www = False


                        # the www directory exists so we copy it to web_dir
                        print('\nCopying web files to {} ...'.format(web_dir))
                        webquiz_util.copytree(webquiz_www, web_dir)

                    self['webquiz_www'] = web_dir
                    files_copied = True

                except PermissionError:
                    print(webquiz_templates.permission_error.format(web_dir))

                except OSError as err:
                    print(webquiz_templates.oserror_copying.format(web_dir=web_dir, err=err))

        if self['webquiz_www']!='SMS':
            # now prompt for the relative url
            webquiz_url = input(webquiz_templates.webquiz_url_message.format(self['webquiz_url']))
            if webquiz_url != '':
                # removing trailing slashes from webquiz_url
                while webquiz_url[-1] == '/':
                    webquiz_url = webquiz_url[:len(webquiz_url) - 1]

                if webquiz_url[0] != '/':  # force URL to start with /
                    webquiz_url = '/' + webquiz_url

                if not web_dir.endswith(webquiz_url):
                    print(webquiz_templates.webquiz_url_warning)
                    input('Press RETURN to continue... ')

                self['webquiz_url'] = webquiz_url

        # save the settings and exit
        self.write_webquizrc()
        print(webquiz_templates.initialise_ending.format(web_dir=self['webquiz_www']))

    def edit_settings(self):
        r'''
        Change current default values for the WebQuiz settings
        '''
        advanced_not_started = True
        for key in self.keys():
            if key not in ['webquiz_www', 'version']:
                if advanced_not_started and self.settings[key]['advanced']:
                    print(webquiz_templates.advanced_settings)
                    advanced_not_started = False

                skey = '{}'.format(self[key])
                setting = input('{}{}[{}]: '.format(
                                    self.settings[key]['help'],
                                    ' ' if len(skey)<40 else '\n',
                                    skey
                          )
                ).strip()
                if setting != '':
                    if key == 'webquiz_url' and setting[0] != '/':
                        print("  ** prepending '/' to webquiz_url **")
                        setting = '/' + setting

                    elif key == 'webquiz_layout':
                        setting = os.path.expanduser(setting)
                        if setting.endswith('.py'):
                            print("  ** removing .py extension from webquiz_layout **")
                            setting = setting[:-3]

                    elif key == 'engine' and setting not in self.settings['engine'].values:
                        print('setting not changed: {} is not a valid TeX engine'.format(setting))
                        setting = self['engine']

                    elif key in ['hide_side_menu', 'random_order']:
                        setting = setting.lower()
                        if setting not in ['true', 'false']:
                            print('setting not changed: {} must be True or False'.format(key))
                            setting = self[key]

                    elif setting=='NONE':
                        setting = ''

                    self[key] = setting

        # save the settings, print them and exit
        self.write_webquizrc()
        self.list_settings()

    def tex_install(self):
        r'''
        Install the tex files into the standard locations in TEXMFMAIN:
            scripts -> TEXMFMAIN/scripts/webquiz
            doc     -> TEXMFMAIN/doc/latex/webquiz
            latex   -> TEXMFMAIN/tex/latex/webquiz
        It is assumed that this is run from the zipfile installation. There
        is little in the way of error checking or debugging.

        Undocumented feature - useful for debugging initialisation routine
        '''
        webquiz_top = os.path.abspath(webquiz_util.webquiz_file('..'))
        texmf = webquiz_util.kpsewhich('-var-value TEXMFMAIN')
        for (src, target) in [('scripts', 'scripts'),
                         ('latex', 'tex/latex'),
                         ('doc', 'doc/latex')]:
            try:
                webquiz_util.copytree(os.path.join(webquiz_top,src), os.path.join(texmf, target, 'webquiz'))

            except (FileExistsError,FileNotFoundError):
                continue

            except PermissionError as err:
                print(webquiz_templates.insufficient_permissions.format(err))
                sys.exit(1)

        try:

            # add a link to webquiz.py
            texbin = os.path.dirname(shutil.which('pdflatex'))
            if sys.platform.startswith('win'):
                shutil.copyfile(
                    os.path.join(texmf,'scripts','webquiz','webquiz.bat'), 
                    os.path.join(texbin, 'webquiz.bat')
                )
            else:
                os.symlink(
                    os.path.join(texmf,'scripts','webquiz','webquiz.py'), 
                    os.path.join(texbin, 'webquiz')
                )
            subprocess.call('mktexlsr', shell=True)

        except (FileExistsError,FileNotFoundError):
            pass

        except PermissionError as err:
            print(webquiz_templates.insufficient_permissions.format(err))
            sys.exit(1)

        except subprocess.CalledProcessError as err:
            self.webquiz_error('There was a problem running mktexlsr', err)

    def tex_uninstall(self):
        r'''
        UnInstall the tex files into TEXMFMAIN. It is assumed that the files
        are installed in the natural locations in the TEXMFMAIN tree, namely:
            scripts -> TEXMFMAIN/scripts/webquiz
            doc     -> TEXMFMAIN/doc/latex/webquiz
            latex   -> TEXMFMAIN/tex/latex/webquiz
        There is little in the way of error checking or debugging.

        Undocumented feature - useful for debugging initialisation routine
        '''
        webquiz_top = os.path.abspath(webquiz_util.webquiz_file('..'))
        texmf = webquiz_util.kpsewhich('-var-value TEXMFMAIN')
        for target in ['scripts', 'tex/latex', 'doc/latex']:
            try:
                shutil.rmtree(os.path.join(texmf, target, 'webquiz'))

            except (FileExistsError,FileNotFoundError):
                pass

            except PermissionError as err:
                print(webquiz_templates.insufficient_permissions.format(err))
                sys.exit(1)

        try:
            # remove link from texbin to webquiz.py
            texbin = os.path.dirname(shutil.which('pdflatex'))
            if sys.platform.startswith('win'):
                os.remove(os.path.join(texbin, 'webquiz.bat'))
            else:
                os.remove(os.path.join(texbin, 'webquiz'))

        except (FileExistsError,FileNotFoundError):
            pass

        except PermissionError as err:
            print(webquiz_templates.insufficient_permissions.format(err))
            sys.exit(1)

        # remove any rcfiles that exist in obvious places
        try:
            if os.path.isfile(self.system_rcfile):
                os.remove(self.system_rcfile)
            if os.path.isfile(self.user_rcfile):
                os.remove(self.user_rcfile)
            if os.path.isfile(self.rcfile):
                os.remove(self.rcfile)
        except PermissionError:
            print(webquiz_templates.insufficient_permissions.format(err))
            sys.exit(1)

        # remove link to webquiz.py
        texbin = os.path.dirname(shutil.which('pdflatex'))
        try:
            if sys.platform.startswith('win'):
                webquiz = os.path.join(texbin, 'webquiz.bat')
                os.remove(webquiz)
            else:
                webquiz = os.path.join(texbin,'webquiz')
                target = os.readlink(webquiz)
                if target==os.path.join(texmf,'scripts','webquiz','webquiz.py'):
                    os.remove(webquiz)

        except (FileExistsError,FileNotFoundError):
            pass

        except OSError as err:
            print('There was a problem removing the link to webquiz: {}'.format(err))

    def uninstall_webquiz(self):
        r'''
        Remove all of the webquiz files from the webserver
        '''

        if os.path.isdir(self['webquiz_www']):
            remove = input('Do you really want to remove the WebQuiz from your web server [N/yes]? ')
            if remove != 'yes':
                print('WebQuiz unistall aborted!')
                return

            try:
                shutil.rmtree(self['webquiz_www'])
                print('WebQuiz files successfully removed from {}'.format(self['webquiz_www']))

            except PermissionError as err:
                print(webquiz_templates.insufficient_permissions.format(err))
                sys.exit(1)

            except OSError as err:
                self.webquiz_error('There was a problem removing webquiz files from {}'.format(self['webquiz_www']), err)

            # now reset and save the locations of the webquiz files and URL
            self['webquiz_url'] = ''
            self['webquiz_www'] = ''
            self.write_webquizrc()

        else:
            self.webquiz_error('uninstall: no webwquiz files are installed on your web server??')

        for rfile in ['system', 'user']:
            rcfile = getattr(self, rfile+'_rcfile')
            if os.path.isfile(rcfile):
                rm = input('Remove {} rcfile {}\n[Y/no] '.format(rfile, rcfile))
                if rm != 'no':
                    try:
                        os.remove(rcfile)
                    except (OSError, PermissionError) as err:
                        self.webquiz_error('There was a problem deleting {}'.format(rcfile), err)


# =====================================================
if __name__ == '__main__':
    try:
        settings = WebQuizSettings()

        # parse the command line options
        parser = argparse.ArgumentParser(description=metadata.description)

        parser.add_argument(
            'quiz_file',
            nargs='*',
            type=str,
            default=None,
            help='latex quiz files')

        parser.add_argument(
            '-q',
            '--quiet',
            action='count',
            default=0,
            help='Suppress tex4ht messages (also -qq etc)')

        parser.add_argument(
            '-d', '--draft',
            action='store_true',
            default=False,
            help='Use make4ht draft mode')

        parser.add_argument(
            '-s',
            '--shell-escape',
            action='store_true',
            default=False,
            help='Shell escape for tex4ht/make4ht')

        engine = parser.add_mutually_exclusive_group()
        engine.add_argument(
            '--latex',
            action='store_const',
            const='latex',
            default=settings['engine'],
            dest='engine',
            help='Use latex to compile document with make4ht (default)')
        engine.add_argument(
            '-l',
            '--lua',
            action='store_const',
            const='lua',
            dest='engine',
            help='Use lualatex to compile the quiz')
        engine.add_argument(
            '-x',
            '--xelatex',
            action='store_const',
            const='xelatex',
            dest='engine',
            help='Use xelatex to compile the quiz')

        parser.add_argument(
            '-r',
            '--rcfile',
            action='store',
            default=None,
            help='Specify location of the webquiz rc-file ')

        settings_parser = parser.add_mutually_exclusive_group()
        settings_parser.add_argument(
            '-i',
            '--initialise',
            '--initialize',
            action='store_true',
            default=False,
            help='Install web components of webquiz')
        settings_parser.add_argument(
            '-e', '--edit-settings',
            action='store_true',
            default=False,
            help='Edit default settings for webquiz')
        settings_parser.add_argument(
            '--settings',
            action='store',
            const='all',
            default='',
            nargs='?',
            type=str,
            help='List default settings for webquiz'
        )
        settings_parser.add_argument(
            '--developer',
            action='store_true',
            default=False,
            help=argparse.SUPPRESS
        )

        # options suppressed from the help message
        parser.add_argument(
            '-m',
            '--make4ht',
            action='store',
            type=str,
            dest='make4ht_options',
            default=settings['make4ht'],
            help=argparse.SUPPRESS
        )

        parser.add_argument(
            '--webquiz_layout',
            action='store',
            type=str,
            dest='webquiz_layout',
            default=settings['webquiz_layout'],
            help=argparse.SUPPRESS
        )

        install_parser = parser.add_mutually_exclusive_group()
        install_parser.add_argument(
            '--tex-install',
            action='store_true',
            default=False,
            help=argparse.SUPPRESS
        )
        install_parser.add_argument(
            '--tex-uninstall',
            action='store_true',
            default=False,
            help=argparse.SUPPRESS
        )
        install_parser.add_argument(
            '--uninstall',
            action='store_true',
            default=False,
            help=argparse.SUPPRESS
        )

        parser.add_argument(
            '--version',
            action='version',
            version='%(prog)s version {}'.format(metadata.version),
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--debugging',
            action='store_true',
            default=False,
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--shorthelp',
            action='store_true',
            default=False,
            help=argparse.SUPPRESS
        )

        # parse the options
        options = parser.parse_args()
        options.prog = parser.prog

        # set debugging mode from options
        settings.debugging = options.debugging

        # read the rcfile and throw an error if we are not adjusting the settings
        if options.rcfile is not None:
            rcfile = os.path.expanduser(options.rcfile)
            settings.read_webquizrc(rcfile)

        if options.uninstall:
            # uninstall web files and exit
            settings.uninstall_webquiz()
            sys.exit()
        elif options.tex_install:
            # install files from zip file into tex distribution and then exit
            settings.tex_install()
            sys.exit()
        elif options.tex_uninstall:
            # install files from zip file into tex distribution and then exit
            settings.tex_uninstall()
            sys.exit()

        # initialise and exit
        if options.initialise or options.developer:
            settings.initialise_webquiz(developer=options.developer)

        # force initialisation if the url is not set
        elif settings['webquiz_url'] == '':
            settings.initialise_webquiz(need_to_initialise=True)

        # list settings and exit
        if options.settings != '':
            settings.list_settings(options.settings)
            sys.exit()

        # edit settings and exit
        if options.edit_settings:
            settings.edit_settings()
            sys.exit()

        # print short help and exit
        if options.shorthelp:
            parser.print_usage()
            sys.exit()

        # if no filename then exit
        if options.quiz_file==[]:
            if settings.have_initialised:
                sys.exit()
            else:
                parser.print_help()
                sys.exit(1)

        # import the local page formatter
        mod_dir, mod_layout = os.path.split(options.webquiz_layout)
        if mod_dir != '':
            sys.path.insert(0, mod_dir)
        options.write_web_page = __import__(mod_layout).write_web_page

        # run() is a shorthand for executing system commands depending on the quietness
        #       - we need to use shell=True because otherwise pst2pdf gives an error
        # options.talk() is a shorthand for letting the user know what is happening
        if options.quiet == 0:
            options.run = webquiz_util.run
            options.talk = lambda msg: print(msg)
        elif options.quiet == 1:
            options.run  = webquiz_util.quiet_run
            options.talk = lambda msg: print(msg)
        else:
            options.run  = webquiz_util.silent_run
            options.talk = lambda msg: None

        # run through the list of quizzes and make them
        for quiz_file in options.quiz_file:
            if len(options.quiz_file) > 1 and options.quiet < 3:
                print('Making web page for {}'.format(quiz_file))

            quiz_name, ext = os.path.splitext(quiz_file)  # extract filename and extension
            # quiz_file is assumed to be a tex file if no extension is given
            if ext=='':
                ext = '.tex'
                quiz_file += ext
 
            # windows likes adding a prefix of '.\\'to filename and this causes havoc with latex
            if os.path.dirname(quiz_file)=='.':
                quiz_file = os.path.basename(quiz_file)

            if ext not in ['.tex', '.xml']:
                    webquiz_util.webquiz_error(True, 'unrecognised file extension {}'.format(ext))

            if not os.path.isfile(quiz_file):
                webquiz_util.webquiz+error(True, 'WebQuiz error: cannot read file {}'.format(quiz_file))

            # the quiz name and the quiz_file will be different if pst2pdf is used
            quiz_name = quiz_file
            if options.quiet < 2:
                print('WebQuiz generating web page for {}'.format(quiz_file))

            options.pst2pdf = False
            if ext==".tex":
                # If the pst2pdf option is used then we need to preprocess
                # the latex file BEFORE passing it to MakeWebQuiz. Set
                # options.pst2pdf = True if pst2pdf is given as an option to
                # the webquiz documentclass
                with codecs.open(quiz_file, 'r', encoding='utf8') as q_file:
                    doc = q_file.read()

                try:
                    brac = doc.index(r'\documentclass[') + 15  # start of class options
                    if 'pst2pdf' in [
                            opt.strip()
                            for opt in doc[brac:brac+doc[brac:].index(']')].split(',')
                    ]:
                        preprocess_with_pst2pdf(options, quiz_file[:-4])
                        options.pst2pdf = True
                        # now run webquiz on the modified tex file
                        quiz_file = quiz_file[:-4] + '-pdf-fixed.tex'
                except ValueError:
                    pass

            # the file exists and is readable so make the quiz
            webquiz_makequiz.MakeWebQuiz(quiz_name, quiz_file, options, settings, metadata)

            quiz_name, ext = os.path.splitext(quiz_name)  # extract filename and extension

            # move the css file into the directory for the quiz
            css_file = os.path.join(quiz_name, quiz_name + '.css')
            if os.path.isfile(quiz_name + '.css'):
                if os.path.isfile(css_file):
                    os.remove(css_file)
                shutil.move(quiz_name + '.css', css_file)

            # now clean up unless debugging
            if not options.debugging:
                for ext in ['4ct', '4tc', 'dvi', 'idv', 'lg', 'log',
                    'ps', 'pdf', 'tmp', 'xml', 'xref'
                ]:
                    if os.path.isfile(quiz_name + '.' + ext):
                        os.remove(quiz_name + '.' + ext)

                # files created when using pst2pdf
                if options.pst2pdf:
                    for file in glob.glob(quiz_name + '-pdf.*'):
                        os.remove(file)
                    for file in glob.glob(quiz_name + '-pdf-fixed.*'):
                        os.remove(file)
                    for extention in ['.preamble', '.plog', '-tmp.tex',
                            '-pst.tex', '-fig.tex'
                    ]:
                        if os.path.isfile(quiz_name + extention):
                            os.remove(quiz_name + extention)
                    if os.path.isdir(os.path.join(quiz_name, quiz_name)):
                        shutil.rmtree(os.path.join(quiz_name, quiz_name))

        if settings.initialise_warning != '' and not settings.have_initialised:
            print(webquiz_templates.text_initialise_warning)

    except Exception as err:

        # there is a small chance that there is an error before the
        # settings.debugging flag has been set
        webquiz_util.webquiz_error(settings.debugging if 'settings' in globals() else True,
            'unknown problem.\n\nIf you think this is a bug please report it by creating an issue at\n    {}\n'
            .format(metadata.repository), err)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024-2025, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


'''
This Python executable is intended for installation within a TeX distribution,
along with Python wheels for the following Python packages:

  * latexminted:  https://pypi.org/project/latexminted/

  * latexrestricted:  https://pypi.org/project/latexrestricted/

  * latex2pydata:  https://pypi.org/project/latex2pydata/

  * Pygments:  https://pypi.org/project/Pygments/

The combined executable plus wheels provide everything that is needed for the
Python side of the minted LaTeX package.  No additional Python libraries are
required.

The wheels require Python >= 3.8.  If this executable is launched with an
earlier Python version, then it will attempt to locate a more recent Python
installation and run itself with that Python version in a subprocess.  The
search for more recent Python versions looks for executables of the form
`python3.x` on PATH.

If the latexminted Python package is installed separately, outside TeX, then
it will create a separate `latexminted` executable as part of the installation
process.  That makes it possible to install latexminted and dependencies
separately and then customize Pygments with additional packages that provide
plugins.  However, running that separate `latexminted` executable is not
straightforward under Windows.  Under Windows, if this executable finds a
suitable `latexminted` executable elsewhere, outside a TeX installation, then
this executable will run that separate `latexminted` executable in a
subprocess and exit.  There are two reasons for this approach:

 1. Under Windows with TeX Live, the default restricted shell escape can only
    run executables such as `latexminted` that are part of the TeX
    installation; the executable that runs is not the first executable on
    PATH.  That is part of TeX Live's security measures to prevent running
    executables in the current working directory, which is typically writable
    by LaTeX and is the first place Windows checks when searching for
    executables.  This script and the latexrestricted Python package enforce
    equivalent security, but do so in a less restrictive manner by expanding
    executable names into executable paths with Python's `shutil.which()` and
    then comparing the result with locations writable by LaTeX.

 2. Under non-Windows operating systems, it is possible to modify PATH so that
    the desired `latexminted` executable is first.  Under Windows, the system
    PATH is prepended to the user PATH, so a system-wide TeX installation will
    prevent a user-installed `latexminted` executable from being accessible.

Requirements for locating and running a separate `latexminted` executable
under Windows:

  * The separate executable must be the first `latexminted.exe` found on PATH,
    or it must be the first `latexminted.exe` on PATH that is located under
    the user home directory.

  * The separate executable must be outside a TeX installation.  There is a
    check for a `tex.exe` executable in the same directory as
    `latexminted.exe`.  There is a check for the case-insensitive strings
    "texlive", "miktex", and "tinytex" in the path to the `latexminted`
    executable.  With TeX Live, the path to the `latexminted` executable is
    also compared to the environment variable `SELFAUTOLOC`.

  * The separate executable must be outside the current working directory,
    TEXMFOUTPUT, and TEXMF_OUTPUT_DIRECTORY.

  * The current directory, TEXMFOUTPUT, and TEXMF_OUTPUT_DIRECTORY cannot be
    subdirectories of the directory in which the executable is located.
'''


# This is the version of the `latexminted.py` launcher script, which is
# bundled with the `minted` package for installation within TeX distributions.
# This is NOT the version of the `latexminted` library.  The library version
# is what matters for compatibility purposes, and the library version is what
# is returned from running `latexminted --version`.
__launcher_version__ = '20250306'


import os
import pathlib
import shutil
import subprocess
import sys




# This is an abbreviated variant of `AnyPath` from latexrestricted:
# https://github.com/gpoore/latexrestricted/blob/main/latexrestricted/_anypath.py
class Path(type(pathlib.Path())):
    __slots__ = (
        '_cache_key',
    )

    if sys.version_info[:2] < (3, 9):
        def is_relative_to(self, other):
            try:
                self.relative_to(other)
                return True
            except ValueError:
                return False

    @property
    def cache_key(self):
        try:
            return self._cache_key
        except AttributeError:
            self._cache_key = (type(self), self)
            return self._cache_key

    _resolved_set = set()

    def resolve(self):
        resolved = super().resolve()
        self._resolved_set.add(resolved.cache_key)
        return resolved

    def is_resolved(self) -> bool:
        return self.cache_key in self._resolved_set




# Define function that determines whether subprocess executable paths are
# permitted.
prohibited_path_roots = set()
prohibited_path_roots.add(Path.cwd())
env_TEXMFOUTPUT = os.getenv('TEXMFOUTPUT')
env_TEXMF_OUTPUT_DIRECTORY = os.getenv('TEXMF_OUTPUT_DIRECTORY')
for env_var in (env_TEXMFOUTPUT, env_TEXMF_OUTPUT_DIRECTORY):
    if env_var:
        env_var_path = Path(env_var)
        prohibited_path_roots.add(env_var_path)
        prohibited_path_roots.add(env_var_path.resolve())

did_init_prohibited_path_roots = False
def is_permitted_executable_path(executable_path, executable_path_resolved):
    global did_init_prohibited_path_roots
    if not did_init_prohibited_path_roots:
        did_init_prohibited_path_roots = True
        init_prohibited_path_roots()
    if not executable_path_resolved.is_resolved():
        raise Exception('Second argument must be resolved path')
    if any(e.is_relative_to(p) or p.is_relative_to(e)
           for e in set([executable_path.parent, executable_path_resolved.parent])
           for p in prohibited_path_roots):
        return False
    return True

# TeX Live allows setting `TEXMFOUTPUT` in LaTeX configuration.
# Retrieving that value with kpsewhich follows the approach in latexrestricted:
# https://github.com/gpoore/latexrestricted/blob/main/latexrestricted/_latex_config.py
env_SELFAUTOLOC = os.getenv('SELFAUTOLOC')
env_TEXSYSTEM = os.getenv('TEXSYSTEM')
def init_prohibited_path_roots():
    if not env_TEXMFOUTPUT and env_SELFAUTOLOC and (not env_TEXSYSTEM or env_TEXSYSTEM.lower() != 'miktex'):
        if sys.platform == 'win32':
            # Under Windows, shell escape executables will often be launched
            # with the TeX Live `runscript.exe` executable wrapper.  This
            # overwrites `SELFAUTOLOC` from TeX with the location of the
            # wrapper, so `SELFAUTOLOC` may not be correct.
            which_tlmgr = shutil.which('tlmgr')  # No `.exe`; likely `.bat`
            if not which_tlmgr:
                sys.exit('Failed to find TeX Live "tlmgr" executable on PATH')
            which_tlmgr_resolved = Path(which_tlmgr).resolve()
            texlive_bin_path = which_tlmgr_resolved.parent
            # Make sure executable is *.exe, not *.bat or *.cmd:
            # https://docs.python.org/3/library/subprocess.html#security-considerations
            which_kpsewhich = shutil.which('kpsewhich.exe', path=str(texlive_bin_path))
            if not which_kpsewhich:
                sys.exit(
                    'Failed to find a TeX Live "tlmgr" executable with accompanying "kpsewhich" executable on PATH'
                )
            which_kpsewhich_path = Path(which_kpsewhich)
            which_kpsewhich_resolved = which_kpsewhich_path.resolve()
            if not texlive_bin_path == which_kpsewhich_resolved.parent:
                sys.exit(' '.join([
                    '"tlmgr" executable from PATH resolved to "{}" '.format(which_tlmgr_resolved.as_posix()),
                    'while "kpsewhich" resolved to "{}";'.format(which_kpsewhich_resolved.as_posix()),
                    '"tlmgr" and "kpsewhich" should be in the same location',
                ]))
            if not which_kpsewhich_resolved.name.lower().endswith('.exe'):
                sys.exit(' '.join([
                    'Executable "kpsewhich" resolved to "{}",'.format(which_kpsewhich_resolved.as_posix()),
                    'but *.exe is required',
                ]))
        else:
            which_kpsewhich = shutil.which('kpsewhich', path=env_SELFAUTOLOC)
            if not which_kpsewhich:
                sys.exit(' '.join([
                    'Environment variable SELFAUTOLOC has value "{}",'.format(env_SELFAUTOLOC),
                    'but a "kpsewhich" executable was not found at that location',
                ]))
            which_kpsewhich_path = Path(which_kpsewhich)
            which_kpsewhich_resolved = which_kpsewhich_path.resolve()
        if not is_permitted_executable_path(which_kpsewhich_path, which_kpsewhich_resolved):
            # As in the latexrestricted case, this doesn't initially check for
            # the TeX Live scenario where `TEXMFOUTPUT` is set in a
            # `texmf.cnf` config file to a location that includes the
            # `kpsewhich` executable.  There isn't a good way to get the value
            # of `TEXMFOUTPUT` without running `kpsewhich` in that case.
            sys.exit(
                'Executable "kpsewhich" is located under the current directory, TEXMFOUTPUT, or '
                'TEXMF_OUTPUT_DIRECTORY, or one of these locations is under the same directory as the executable'
            )
        kpsewhich_cmd = [which_kpsewhich_resolved.as_posix(), '--var-value', 'TEXMFOUTPUT']
        try:
            kpsewhich_proc = subprocess.run(kpsewhich_cmd, shell=False, capture_output=True)
        except PermissionError:
            sys.exit('Insufficient permission to run "{}"'.format(which_kpsewhich_resolved.as_posix()))
        kpsewhich_TEXMFOUTPUT = kpsewhich_proc.stdout.strip().decode(sys.stdout.encoding) or None
        if kpsewhich_TEXMFOUTPUT:
            kpsewhich_TEXMFOUTPUT_path = Path(kpsewhich_TEXMFOUTPUT)
            prohibited_path_roots.add(kpsewhich_TEXMFOUTPUT_path)
            prohibited_path_roots.add(kpsewhich_TEXMFOUTPUT_path.resolve())
        if not is_permitted_executable_path(which_kpsewhich_path, which_kpsewhich_resolved):
            # It is now possible to check for the TeX Live scenario where
            # `TEXMFOUTPUT` is set in a `texmf.cnf` config file to a location
            # that includes the `kpsewhich` executable.  Giving an error
            # message after already running `kpsewhich` isn't ideal, but there
            # isn't a good alternative.  As in the latexrestricted case, the
            # impact on overall security is negligible because an unsafe value
            # of `TEXMFOUTPUT` means that all TeX-related executables are
            # potentially compromised.
            sys.exit(
                'Executable "kpsewhich" is located under the current directory, TEXMFOUTPUT, or '
                'TEXMF_OUTPUT_DIRECTORY, or one of these locations is under the same directory as the executable'
            )




# If Python version is < 3.8, try to locate a more recent version and then
# relaunch this script with that Python version in a subprocess.
if sys.version_info[:2] < (3, 8):
    for minor_version in range(13, 7, -1):
        if sys.platform == 'win32':
            # Batch files must be prohibited:
            # https://docs.python.org/3/library/subprocess.html#security-considerations
            which_python = shutil.which('python3.{}.exe'.format(minor_version))
        else:
            which_python = shutil.which('python3.{}'.format(minor_version))
        if which_python:
            which_python_path = Path(which_python)
            which_python_resolved = which_python_path.resolve()
            if sys.platform == 'win32' and not which_python_resolved.name.lower().endswith('.exe'):
                continue
            if is_permitted_executable_path(which_python_path, which_python_resolved):
                python_cmd = [which_python_resolved.as_posix(), __file__] + sys.argv[1:]
                python_proc = subprocess.run(python_cmd, shell=False, capture_output=True)
                sys.stderr.buffer.write(python_proc.stderr)
                sys.stdout.buffer.write(python_proc.stdout)
                sys.exit(python_proc.returncode)
    sys.exit('latexminted requires Python >= 3.8, but a compatible Python executable was not found on PATH')




# Check for required wheel dependencies and add them to Python's `sys.path`.
script_resolved = Path(__file__).resolve()
required_wheel_packages = (
    'latexminted',
    'latexrestricted',
    'latex2pydata',
    'pygments',
)
wheel_paths = [p for p in script_resolved.parent.glob('*.whl') if p.name.startswith(required_wheel_packages)]
if not wheel_paths:
    sys.exit('latexminted failed to find bundled wheels *.whl')
for pkg in required_wheel_packages:
    if not any(whl.name.startswith(pkg) for whl in wheel_paths):
        sys.exit('latexminted failed to find all required bundled wheels *.whl')
for wheel_path in wheel_paths:
    sys.path.insert(0, wheel_path.as_posix())




# Under Windows, check PATH for a `latexminted` executable outside a TeX
# installation.  If a `latexminted` executable is found in a suitable location
# with sufficient precedence, run it in a subprocess and exit.
#
# The environment variable `LATEXMINTED_SUBPROCESS` is used to prevent an
# endless recursion of subprocesses in the event that a `latexminted`
# executable *inside* a TeX installation somehow manages to pass the tests for
# an executable *outside* a TeX installation.
if sys.platform == 'win32' and not os.getenv('LATEXMINTED_SUBPROCESS'):
    os.environ['LATEXMINTED_SUBPROCESS'] = '1'
    fallback_path_search = True
    if env_SELFAUTOLOC:
        env_SELFAUTOLOC_resolved = Path(env_SELFAUTOLOC).resolve()
    else:
        env_SELFAUTOLOC_resolved = None
    which_latexminted = shutil.which('latexminted.exe')
    if which_latexminted:
        which_latexminted_path = Path(which_latexminted)
        which_latexminted_resolved = which_latexminted_path.resolve()
        if not which_latexminted_resolved.name.lower().endswith('.exe'):
            sys.exit(' '.join([
                'Executable "latexminted" resolved to "{}",'.format(which_latexminted_resolved.as_posix()),
                'but *.exe is required',
            ]))
        if which_latexminted_resolved == script_resolved:
            pass
        elif (which_latexminted_resolved.parent / 'tex.exe').exists():
            pass
        elif any(x in which_latexminted_resolved.as_posix().lower() for x in ('texlive', 'miktex', 'tinytex')):
            pass
        elif env_SELFAUTOLOC_resolved and which_latexminted_resolved.is_relative_to(env_SELFAUTOLOC_resolved):
            pass
        elif is_permitted_executable_path(which_latexminted_path, which_latexminted_resolved):
            latexminted_cmd = [which_latexminted_resolved.as_posix()] + sys.argv[1:]
            latexminted_proc = subprocess.run(latexminted_cmd, shell=False, capture_output=True)
            sys.stderr.buffer.write(latexminted_proc.stderr)
            sys.stdout.buffer.write(latexminted_proc.stdout)
            sys.exit(latexminted_proc.returncode)
        else:
            # If there was a `latexminted` executable on PATH outside a TeX
            # installation, but it wasn't permitted due to its location, don't
            # perform fallback search.
            fallback_path_search = False
    if fallback_path_search:
        # Windows appends user PATH to system PATH, so the system PATH may
        # prevent finding a user installation of `latexminted`.  Search
        # through PATH elements under user home directory to check for
        # `latexminted.exe` outside a TeX installation.
        home_path = Path.home()
        env_PATH = os.environ.get('PATH', '')
        for path_elem in env_PATH.split(os.pathsep):
            if not path_elem or not Path(path_elem).is_relative_to(home_path):
                continue
            which_latexminted = shutil.which('latexminted.exe', path=path_elem)
            if not which_latexminted:
                continue
            which_latexminted_path = Path(which_latexminted)
            which_latexminted_resolved = which_latexminted_path.resolve()
            if which_latexminted_resolved == script_resolved:
                break
            elif (which_latexminted_resolved.parent / 'tex.exe').exists():
                break
            elif any(x in which_latexminted_resolved.as_posix().lower() for x in ('texlive', 'miktex', 'tinytex')):
                break
            elif env_SELFAUTOLOC_resolved and which_latexminted_resolved.is_relative_to(env_SELFAUTOLOC_resolved):
                break
            elif is_permitted_executable_path(which_latexminted_path, which_latexminted_resolved):
                latexminted_cmd = [which_latexminted_resolved.as_posix()] + sys.argv[1:]
                try:
                    latexminted_proc = subprocess.run(latexminted_cmd, shell=False, capture_output=True)
                except PermissionError:
                    break
                sys.stderr.buffer.write(latexminted_proc.stderr)
                sys.stdout.buffer.write(latexminted_proc.stdout)
                sys.exit(latexminted_proc.returncode)
            else:
                break




from latexminted.cmdline import main
main()



r'''
-----------------------------------------------------------------------------
    webquiz_local | default layout of webquiz quiz pages.
-----------------------------------------------------------------------------

    Copyright (C) Andrew Mathas, University of Sydney

    Distributed under the terms of the GNU General Public License (GPL)
                  http://www.gnu.org/licenses/

    This file is part of the WebQuiz system.

    <Andrew.Mathas@sydney.edu.au>
-----------------------------------------------------------------------------
'''

# -*- encoding: utf-8 -*-

from webquiz_templates import no_script

# If you want to create a new version of this file to change the
# quiz web page layout then the following "dictionary" of code variables
# is likely to be useful:
#   quiz.breadcrumbs = HTML for breadcrunbs
#   quiz.css = css includes for quisz
#   quiz.department = department from \UnitDepartment
#   quiz.department_url = unit url from \UnitURL
#   quiz.header = header links for quiz
#   quiz.institution = institution from \Institution
#   quiz.institution_url = unit url from \InstitutionURL
#   quiz.javascript = javascript includes for quiz
#   quiz.quiz_header = HTMK for quiz title and navigation arrows
#   quiz.quiz_questions = html for quiz
#   quiz.side_menu = HTML for side menu, including navigation buttons
#   quiz.title = web page title from \title{...} command
#   quiz.unit_code = unit code from \UnitCode
#   quiz.unit_name = unit name from \UnitName
#   quiz.unit_url = unit url from \UnitURL
#   quiz.webquiz_init = javascript for initialising quiz page
#                       MUST appear towards the end of HTML body

def write_web_page(quiz):
  return quiz_page.format(
    title=quiz.quiz.title,  # page title
    htmlpreamble=quiz.header + quiz.javascript + quiz.css,  # header material
    breadcrumbs=quiz.breadcrumbs,  # bread crumb constructed above
    side_menu=quiz.side_menu,  # navigation menu for quiz
    quiz_header=quiz.quiz_header,  # quiz title + navigation arrows
    quiz_questions=quiz.quiz_questions,  # html for quiz
    no_script=no_script,  # error when javascript is not enabled
    webquiz_init=quiz.webquiz_init  # parting javascript callsWebQuizInt
  )


quiz_page = r'''<!DOCTYPE HTML>
<html lang="en">
<head>
  <title> {title} </title>
  {htmlpreamble}
</head>
<body>
  {no_script}{breadcrumbs}
  <div class="quiz-page">
    {side_menu}
    <div class="quiz-questions">
      {quiz_header}
      {quiz_questions}
    </div>
  </div>
  {webquiz_init}
</body>
</html>
'''



#!/bin/python3
"""
A hack to allow decoding for example ^^[ to the proper escape sequence.
"""

import sys
import re
def replace_8bit(match_: re.Match) -> str:
	return chr(ord(match_.group(1))^0x40)
if __name__ == '__main__':
	for line in sys.stdin:
		sys.stdout.write(re.sub(r"\^\^(.)", replace_8bit, line))



#!/usr/bin/python
"""
    pymergechanges: Merge commits made with changes package into text
    Copyright (C) 2018  Y. Cui

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

pymergechanges: Merge commits made with changes package into text

Requires Python version 3; Tested on Python 3.6.6

Supported commands: added, deleted, replaced, highlight
Usage: python pyMergeChanges.py [-arh] <Input File> <Output File>
<Output File> will be overwritten and must be different than <Input File>.
Options:
    -a: accept all added, deleted and replaced
    -r: reject all added, deleted and replaced
    -h: remove all highlights
If no option is given, runs interactively.

Created on Wed Dec  5 20:28:40 2018
Revised on Tue Aug 27 17:51:58 2019

"""

import sys
import re
import codecs

def parse_param(parstr):
    if all(p in parstr for p in ('a', 'r')):
        print('You cannot accept and reject at the same time.')
        sys.exit(1)
    parout = ''
    for par in parstr[1:]:
        if par == 'a':
            print('Accepting all added, deleted and replaced')
            parout += 'a'
        elif par == 'r':
            print('Rejecting all added, deleted and replaced')
            parout += 'r'
        elif par == 'h':
            print('Removing all highlights.')
            parout += 'h'
        else:
            print('Unknown parameter: ' + par)
            sys.exit(1)
    if not parout:
        parout = 'i'
    return parout

def ask1(): # accept, reject, keep, break
    if PARAMS == 'i':
        while True:
            ans = input('[a]ccept or [r]eject or [k]eep or [b]reak ? ').lower()
            if ans == 'a':
                print('Accepted.')
                break
            elif ans == 'r':
                print('Rejected.')
                break
            elif ans == 'k':
                print('Kept.')
                break
            elif ans == 'b':
                print('Alright.')
                break
    elif any(p == 'a' for p in PARAMS):
        print('Accepted.')
        ans = 'a'
    elif any(p == 'r' for p in PARAMS):
        print('Rejected.')
        ans = 'r'
    else:
        ans = 'k'
    return ans

def ask2(): # remove, keep, break
    if PARAMS == 'i':
        while True:
            ans = input('[r]emove or [k]eep or [b]reak ? ').lower()
            if ans == 'r':
                print('Removed.')
                break
            elif ans == 'k':
                print('Kept.')
                break
            elif ans == 'b':
                print('Alright.')
                break
        return ans
    elif any(p == 'h' for p in PARAMS):
        print('Removed.')
        ans = 'r'
    else:
        print('Kept.')
        ans = 'k'
    return ans

def trim_space(line, pos): # trim two consecutive white spaces
    if line[pos-1:pos+1] == '  ':
        line = line[:pos] + line[pos+1:]
    return line

# BEGIN OF SCRIPT

if len(sys.argv) not in [3, 4]:
    print(__doc__)
    sys.exit(1)

# parse input parameters
if len(sys.argv) == 3:
    print('Running in interactive mode. ')
    INPUTFILE, OUTPUTFILE = sys.argv[1:]
    PARAMS = "i"
if len(sys.argv) == 4:
    PARAMLIST, INPUTFILE, OUTPUTFILE = sys.argv[1:]
    PARAMS = parse_param(PARAMLIST)

if INPUTFILE == OUTPUTFILE:
    print('Input File and Output File must be different.')
    sys.exit(1)

RE_ADDED = re.compile(r'(\\added)(\[[^\]]*\])?\{(([^\}]?(\{[^\}]*\})?)*)\}')
RE_DELETED = re.compile(r'(\\deleted)(\[[^\]]*\])?\{(([^\}]?(\{[^\}]*\})?)*)\}')
RE_REPLACED = re.compile(r'(\\replaced)(\[[^\]]*\])?\{(([^\}]?(\{[^\}]*\})?)*)\}\{(([^\}]?(\{[^\}]*\})?)*)\}')
RE_HIGHLIGHT = re.compile(r'(\\highlight)(\[[^\]]*\])?\{(([^\}]?(\{[^\}]*\})?)*)\}')
RE_COMMENT = re.compile(r'(\\comment)(\[[^\]]*\])?\{(([^\}]?(\{[^\}]*\})?)*)\}')

codecs.open(OUTPUTFILE, mode='w', encoding='utf8').close()
with codecs.open(INPUTFILE, mode='r', encoding='utf8') as fin, \
codecs.open(OUTPUTFILE, mode='a', encoding='utf8') as fout:
    LINE_COUNT = 0
    FLAG_FAST_BREAK = False # if you want to halt merging for some reason
    for linein in fin:
        if FLAG_FAST_BREAK:
            fout.write(linein)
            continue
        LINE_COUNT += 1
        lineout = linein
        matchAdded = RE_ADDED.search(lineout)
        matchDeleted = RE_DELETED.search(lineout)
        matchReplaced = RE_REPLACED.search(lineout)
        matchHighlight = RE_HIGHLIGHT.search(lineout)
        matchComment = RE_COMMENT.search(lineout)
        flagHasCommit = matchAdded or matchDeleted or matchReplaced or matchHighlight or matchComment
        if flagHasCommit:
            print('\n******** In Line %i :\n %s' % (LINE_COUNT, lineout))
            next_pos = 0
            while matchAdded:
                if FLAG_FAST_BREAK:
                    break
                print('\n** add commit ** \n' + matchAdded.group())
                answer = ask1()
                if answer == 'a':
                    lineout = (lineout[:matchAdded.start(0)]
                               + matchAdded.group(3) + lineout[matchAdded.end(0):])
                    lineout = trim_space(lineout, matchAdded.start(0)
                                         + len(matchAdded.group(3)))
                    lineout = trim_space(lineout, matchAdded.start(0))
                elif answer == 'r':
                    lineout = (lineout[:matchAdded.start(0)] + lineout[matchAdded.end(0):])
                    lineout = trim_space(lineout, matchAdded.start(0))
                elif answer == 'k':
                    lineout = lineout
                    next_pos = matchAdded.end(0)
                elif answer == 'b':
                    FLAG_FAST_BREAK = True
                    break
                matchAdded = RE_ADDED.search(lineout, next_pos) # redo matching for updated text

            next_pos = 0
            matchDeleted = RE_DELETED.search(lineout, next_pos)  # redo matching for (potentially) updated text
            while matchDeleted:
                if FLAG_FAST_BREAK:
                    break
                print('\n** delete commit ** \n' + matchDeleted.group())
                answer = ask1()
                if answer == 'a':
                    lineout = (lineout[:matchDeleted.start(0)] + lineout[matchDeleted.end(0):])
                    lineout = trim_space(lineout, matchDeleted.start(0))
                elif answer == 'r':
                    lineout = (lineout[:matchDeleted.start(0)] + matchDeleted.group(3)
                               + lineout[matchDeleted.end(0):])
                elif answer == 'k':
                    lineout = lineout
                    next_pos = matchDeleted.end(0)
                elif answer == 'b':
                    FLAG_FAST_BREAK = True
                    break
                matchDeleted = RE_DELETED.search(lineout, next_pos)

            next_pos = 0
            matchReplaced = RE_REPLACED.search(lineout, next_pos)  # redo matching for (potentially) updated text
            while matchReplaced:
                if FLAG_FAST_BREAK:
                    break
                print('\n** replace commit ** \n' + matchReplaced.group())
                answer = ask1()
                if answer == 'a':
                    lineout = (lineout[:matchReplaced.start(0)]
                               + matchReplaced.group(3) + lineout[matchReplaced.end(0):])
                    lineout = trim_space(lineout, matchReplaced.start(0)
                                         + len(matchReplaced.group(3)))
                    lineout = trim_space(lineout, matchReplaced.start(0))
                elif answer == 'r':
                    lineout = (lineout[:matchReplaced.start(0)]
                               + matchReplaced.group(6) + lineout[matchReplaced.end(0):])
                    lineout = trim_space(lineout, matchReplaced.start(0)
                                         + len(matchReplaced.group(6)))
                    lineout = trim_space(lineout, matchReplaced.start(0))
                elif answer == 'k':
                    lineout = lineout
                    next_pos = matchReplaced.end(0)
                elif answer == 'b':
                    FLAG_FAST_BREAK = True
                    break
                matchReplaced = RE_REPLACED.search(lineout, next_pos)

            next_pos = 0
            matchHighlight = RE_HIGHLIGHT.search(lineout, next_pos)  # redo matching for (potentially) updated text
            while matchHighlight:
                if FLAG_FAST_BREAK:
                    break
                print('\n** highlight commit ** \n' + matchHighlight.group())
                answer = ask2()
                if answer == 'r':
                    lineout = (lineout[:matchHighlight.start(0)]
                               + matchHighlight.group(3) + lineout[matchHighlight.end(0):])
                    lineout = trim_space(lineout, matchHighlight.start(0)
                                         + len(matchHighlight.group(3)))
                    lineout = trim_space(lineout, matchHighlight.start(0))
                elif answer == 'k':
                    lineout = lineout
                    next_pos = matchHighlight.end(0)
                elif answer == 'b':
                    FLAG_FAST_BREAK = True
                    break
                matchHighlight = RE_HIGHLIGHT.search(lineout, next_pos)

            next_pos = 0
            matchComment = RE_COMMENT.search(lineout, next_pos)  # redo matching for (potentially) updated text
            while matchComment:
                if FLAG_FAST_BREAK:
                    break
                print('\n** comment commit ** \n' + matchComment.group())
                answer = ask2()
                if answer == 'r':
                    lineout = (lineout[:matchComment.start(0)] + lineout[matchComment.end(0):])
                    lineout = trim_space(lineout, matchComment.start(0))
                elif answer == 'k':
                    lineout = lineout
                    next_pos = matchComment.end(0)
                elif answer == 'b':
                    FLAG_FAST_BREAK = True
                    break
                matchComment = RE_COMMENT.search(lineout, next_pos)

            if lineout.isspace():
                print('\nResult is empty line and not stored.')
            else:
                print('\nResult: \n' + lineout + '\n')
                fout.write(lineout)

        else:
            fout.write(linein)
#
# END OF SCRIPT



r'''
-----------------------------------------------------------------------------
    webquiz_xml | xml reader for reading the xml file generated by tex4ht
-----------------------------------------------------------------------------

    Copyright (C) Andrew Mathas and Donald Taylor, University of Sydney

    Distributed under the terms of the GNU General Public License (GPL)
                  http://www.gnu.org/licenses/

    This file is part of the Math_quiz system.

    <Andrew.Mathas@sydney.edu.au>
    <Donald.Taylor@sydney.edu.au>
-----------------------------------------------------------------------------
'''

# -*- encoding: utf-8 -*-

import xml.sax

# imports of webquiz code
import webquiz_util

# ---------------------------------------------------------------------------------------
def ReadWebQuizXmlFile(quizfile, defaults):
    r'''
    Set up, call and then return the xml parser
    for the quiz web page
    '''
    parser = xml.sax.make_parser()
    quiz = QuizHandler(defaults)
    parser.setContentHandler(quiz)
    parser.setErrorHandler(quiz)
    parser.setDTDHandler(quiz) # as far as I can see this does nothing...
    parser.parse(quizfile)
    parser.close()
    return quiz


# ---------------------------------------------------------------------------------------
class Data(object):
    r'''
    A wrapper object class that holds the data for the different
    components of the quiz.
    '''
    def __init__(self, **args):
        '''
        Accepts key-value pairs, each of which is stored as an attribute
        '''
        self._items = args.items()
        for key, val in args.items():
            setattr(self, key, val)


class QuizHandler(xml.sax.ContentHandler):
    """
        The content handler gives the xml tags to `startElement`, which
        initialises the webquiz tags, and then `endElement` attaches the
        content of each webquiz tag tothe appropriate part of `self`. Any end
        tag that is ot special to webquiz has its contents appended to
        `self.text`. Any tag that contains `DeFaUlT` is set to the system
        default using the `defaults` dictionary.
    """

    def __init__(self, defaults):
        self.defaults = defaults

        # arrays for the different quiz components
        self.discussion_list = []
        self.question_list = []
        self.quiz_index = []

        # these will contain the link and meta elements from the xml file but they
        # are ignored by webquiz.py
        self.link_list = []
        self.meta_list = []

        # to add mathjs when an eval comparison is used
        self.mathjs = False

        # the following tags have defaults set by `defaults`
        self.setting_tags = [
               'department',
               'department_url',
               'institution',
               'institution_url',
               'language',
               'theme',
        ]
        # quiz data
        for tag in self.setting_tags:
            setattr(self, tag, defaults[tag])
        self.breadcrumb = ''
        self.text = ''
        self.after_text = ''
        self.title = ''
        self.unit_code = ''
        self.unit_name = ''

        # keep track of current tags for debugging...
        self.current_tags=[]

    def webquiz_debug(self, msg):
        r'''
            Customised debugging message for the xml module
        '''
        webquiz_util.webquiz_debug(self.defaults.debugging, 'xml: '+msg)

    def webquiz_error(self, msg, err=None):
        r'''
            Customised error message for the xml module
        '''
        webquiz_util.webquiz_error(self.defaults.debugging, 'xml: '+msg, err)

    def set_default_attribute(self, key, value):
        ''' Set the attribute `key` of self, using the default value if
        `value=='DeFaUlT'`.
        '''
        if value.strip() == 'DeFaUlT':
            setattr(self, key, self.defaults[key])
        else:
            setattr(self, key, value)
        self.webquiz_debug('Just set "{}" equal to "{}" from "{}"'.format(key, getattr(self, key), value))

    #---- start of start elements --------------------------------------------
    def startElement(self, tag, attributes):
        '''
            At the start of each webquiz xml tag we need to pull out the
            attributes and place
        '''
        self.webquiz_debug('Starting tag for '+tag)
        self.current_tags.append(tag)

        if hasattr(self, 'start_'+tag):
            getattr(self, 'start_'+tag)(attributes)

        elif tag in ['department', 'institution', 'uni']:
            for key in attributes.keys():
                self.set_default_attribute(tag, attributes.get(key))

    def start_webquiz(self, attributes):
        r'''
        Start element for tag="webquiz". Initialise the quix and extract
        and process the list of attributes.
        '''
        for key in attributes.keys():
            self.set_default_attribute(key, attributes.get(key))

        # convert the following attibutes to booleans
        for key in ['debugging', 'hide_side_menu', 'one_page', 'pst2pdf', 'random_order']:
            setattr(self, key, getattr(self, key)=='true')

        setattr(self, 'language', self.language.lower())
        setattr(self, 'theme', self.theme.lower())

        # set debugging mode from the latex file...from this point on
        self.defaults.debugging = self.defaults.debugging or self.debugging

    def start_link(self, attributes):
        r'''
        Start element for tag="link"
        '''
        self.link_list.append({key: attributes.get(key) for key in attributes.keys()})

    def start_meta(self, attributes):
        r'''
        Start element for tag="meta"
        '''
        self.meta_list.append({key: attributes.get(key) for key in attributes.keys()})

    def start_breadcrumb(self, attributes):
        r'''
        Start element for tag="breadcrumb"
        '''
        self.set_default_attribute('breadcrumbs', attributes.get('breadcrumbs'))

    def start_unit_name(self, attributes):
        r'''
        Start element for tag="unit_name"
        '''
        self.set_default_attribute('unit_url', attributes.get('url'))
        self.quizzes_url = attributes.get('quizzes_url')
        if self.quizzes_url == 'DeFaUlT':
            self.quizzes_url = self.unit_url + '/Quizzes'

    def start_discussion(self, attributes):
        r'''
        Start element for tag="discussion"
        '''
        discussion = Data(heading = '',
                          short_heading = '',
                          text = ''  # The text of the discussion
        )
        self.discussion_list.append(discussion)

    def start_question(self, attributes):
        r'''
        Start element for tag="question"
        '''
        self.question_list.append(
            Data(text = '',      # The text of the question
                type = None,    # input, or single or multiple choice
                after_text = '' # text at end of question
            )
        )

    def start_answer(self, attributes):
        r'''
        Process the different question types, items choice and feedback
        '''
        if self.question_list[-1].type != None:
            self.webquiz_error('question {} has too many question types: {} and input'.format(
                    len(self.question_list)+1, self.question_list[-1].type)
            )
        self.question_list[-1].type = 'input'
        self.question_list[-1].answer = ''
        self.question_list[-1].feedback_right = ''
        self.question_list[-1].feedback_wrong = ''
        self.question_list[-1].text += self.text
        self.text = ''

        self.question_list[-1].comparison = attributes.get('comparison')
        self.question_list[-1].prompt = attributes.get('prompt')=='true'
        if self.question_list[-1].comparison in ['complex', 'number']:
            self.mathjs = True

    def start_choice(self, attributes):
        r'''
        Start element for tag="choice"
        '''
        if self.question_list[-1].type != None:
            self.webquiz_error('question {} has too many question types: {} and choice'.format(
                    len(self.question_list)+1, self.question_list[-1].type)
            )
        self.question_list[-1].type = attributes.get('type')
        self.question_list[-1].columns = int(attributes.get('columns'))
        self.question_list[-1].items = []
        self.question_list[-1].correct = 0
        self.question_list[-1].text += self.text
        self.text = ''

    def start_item(self, attributes):
        r'''
        Start element for tag="item"
        '''
        self.question_list[-1].items.append(
                Data(correct= attributes.get('correct'),
                     symbol=attributes.get('symbol'),
                     feedback='',
                     text=''
                    )
        )
        if attributes.get('correct')=='true':
            self.question_list[-1].correct += 1

    def start_index_item(self, attributes):
        r'''
        Finally look after the index file
        '''
        self.quiz_index.append(Data(
                prompt=attributes.get('prompt')=='true',
                url=attributes.get('url'),
                title=''
            )
        )

    def start_when(self, attributes):
        r'''
        start element for tag="when"
        '''
        if self.text.strip() != '':
            self.question_list[-1].after_text += ' '+self.text.strip()
            self.webquiz_debug('After_text is now {}'.format(self.question_list[-1].after_text))
            self.text = ''
        self.current_tags[-1] = 'feedback_'+attributes.get('type')

    #---- end of start elements ---------------------------------------------

    def endElement(self, tag):
        self.webquiz_debug('ending tag for {} (should be {})'.format(tag, self.current_tags[-1])) 

        reset_text = True
        if hasattr(self, 'end_'+tag):
            getattr(self, 'end_'+tag)()
            self.text = ''

        elif tag in self.setting_tags:
            self.set_default_attribute(tag, self.text)
            self.text = ''

        elif tag in ['heading', 'short_heading']:
            setattr(self.discussion_list[-1], tag, self.text.strip())

        elif tag in ['breadcrumb', 'title', 'unit_code', 'unit_name']:
            setattr(self, tag, self.text.strip())

        else:
            # self.text lives to be used another day
            reset_text = False

        if reset_text:
            self.text = ''

        # remove the last tag from the tag list
        self.current_tags.pop()

    #---- start of the end elements ------------------------------------------

    def end_answer(self):
        r'''
        Process end tag when tag="answer"
        '''
        self.question_list[-1].answer = self.text.strip()

    def end_discussion(self):
        r'''
        Process end tag when tag="discussion"
        '''
        self.discussion_list[-1].text = self.text.strip()

    def end_item(self):
        r'''
        Process end tag when tag="item"
        '''
        self.question_list[-1].items[-1].text = self.text.strip()

    def end_feedback(self):
        r'''
        Process end tag when tag="feedback"
        '''
        self.question_list[-1].items[-1].feedback = self.text.strip()

    def end_question(self):
        r'''
        Process end tag when tag="question"
        '''
        # first some error checking
        if self.question_list[-1].type == None:
                self.webquiz_error('Question {} does not have an \\answer or choice environment'.format(
                              len(self.question_list)+1))

        elif hasattr(self.question_list[-1], 'items'):
            if len(self.question_list[-1].items)==0:
                self.webquiz_error('question {} has no multiple choice items'.format(
                              len(self.question_list)+1))

            if self.question_list[-1].type=='single' and self.question_list[-1].correct!=1:
                self.webquiz_error('question {} is single-choice but has {} correct answers'.format(
                                len(self.question_list)+1,
                                self.question_list[-1].correct
                             )
                )
        elif not hasattr(self.question_list[-1], 'answer') or self.question_list[-1].answer=='':
            self.webquiz_error('question {} does have not an \answer or multiple choice'.format(
                          len(self.question_list)+1))

        if self.text.strip() != '':
            self.question_list[-1].after_text += ' '+self.text.strip()

    def end_index_item(self):
        r'''
        Process end tag when tag="index_item"
        '''
        self.quiz_index[-1].title = self.text.strip().replace('\n',' ').replace('\r',' ')

    def end_when(self):
        r'''
        Process end tag when tag="index_item"
        '''
        self.webquiz_debug('WHEN: Adding text to '+self.current_tags[-1])
        setattr(self.question_list[-1], self.current_tags[-1], self.text.strip())

    #---- end of end elements -----------------------------------------------

    def characters(self, text):  #data,start,length):
        r'''
        Append everything to `self.text`
        '''
        self.text += text

    def error(self, e):
        self.webquiz_error('unknown error', e)

    def fatalError(self, e):
        self.webquiz_error('unknown fatal error', e)



r'''
-----------------------------------------------------------------------
    webquiz_templates | html template file
-----------------------------------------------------------------------

    Copyright (C) Andrew Mathas, University of Sydney

    Distributed under the terms of the GNU General Public License (GPL)
              http://www.gnu.org/licenses/

    This file is part of the WebQuiz system.

    <Andrew.Mathas@sydney.edu.au>
----------------------------------------------------------------------
'''

# -*- encoding: utf-8 -*-

## The quiz web pages are built using the following "template" strings

# html meta statements
html_meta = r'''<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
  <meta name="viewport" content="width=device-width,initial-scale=1.0">
  <meta name="generator" content="WebQuiz {version} (http://www.maths.usyd.edu.au/u/mathas/WebQuiz/webquiz-online-manual.html)">
  <meta name="description" content="{description}">
  <meta name="authors" content="WebQuiz: {authors}">
  <meta name="organization" content="{department}, {institution}">
  <meta name="Copyright" content="WebQuiz: {copyright}">
  <meta name="keywords" content="WebQuiz, TeX4ht, make4ht, latex, python, quiz, mathematics">
  <link href="{webquiz_url}/css/webquiz-{theme}.css" type="text/css" rel="stylesheet">
  <link href="{quiz_file}/{quiz_file}.css" type="text/css" rel="stylesheet">
'''

# javascript for setting up the questions
questions_javascript = r'''  <script src="{webquiz_url}/js/webquiz.js"></script>
  <script defer src="{mathjax}?config=MML_CHTML"></script>'''

mathjs=r'  <script defer src="https://cdnjs.cloudflare.com/ajax/libs/mathjs/5.4.0/math.min.js"></script>'

webquiz_init = r'''<div style="display: none;">
    <script src="quizindex.js"></script>
    <script>WebQuizInit({number_questions}, {number_discussions}, '{quiz_file}');</script>
  </div>'''

# Bread crumbs including a drop down menu for all of the quizzes for the unit.
# The drop-down-menu is added by create_quizindex_menu() in webquiz.js
breadcrumb_line_text = '            <li>{text}</li>\n'
breadcrumb_line_url  = '            <li><a href="{url}">{text}</a></li>\n'
breadcrumb_quizindex  = r'''              <li><a href="{quizzes_url}">{quizzes}</a>
                  <span onclick="toggle_quizindex_menu();" id="quizzes-menu-icon"></span>
                  <ul id="quizindex-menu" onclick="toggle_quizindex_menu();"></ul>
              </li>
'''
create_quizindex_menu = r'''// construct the drop down menu if QuizTitles has some entries
if (QuizTitles.length > 0 && quizindex_menu) {
    create_quizindex_menu();
}
'''
breadcrumbs = r'''<div class="breadcrumbs">
    <nav>
        <div class="navleft">
            <ul>{crumbs}            </ul>
        </div>
    </nav>
  </div>
'''

# Should we add a menu to change the theme dynamically? If so then the code
# below should be added to the breadcrumbs and then fixed a little. Perhaps the
# trickiest bit is finding the current list of supported themes.
theme_menu=r'''
        <div class="navright" >
            <span onclick="toggle_theme_menu();">Theme &#9881;</span>
            <ul><li>
                <ul id="theme-menu" onclick="toggle_theme_menu();">
                  <li>one</li><li>two</li>
                </ul>
              </li>
            </ul>
        </div>
'''

# Add a drop-down menu to the navigation to dynamically change the theme?

# question buttons
button = r'        <div id="button{b}" class="button {cls}" content=" " onClick="gotoQuestion({b})">{b}</div>'
discuss = r'        <li id="button-{b}" class="discussion" onClick="gotoQuestion(-{b})">{title}</li>'
side_menu = r'''<div id="menu-icon">
      <span id="sidelabelclosed" class="question-label" onclick="toggle_side_menu();">&#10070;</span>
      <span id="sidelabelopen" class="question-label" onclick="toggle_side_menu();">&#10006;&nbsp;{side_questions}
      </span>
    </div>
    <div id="sidemenu" class="side-menu">{discussion_list}{question_buttons}
      <div class="school">
        {department}<p>
        {institution}
      </div>
      <div class="copyright">
        <a href="http://www.maths.usyd.edu.au/u/mathas/WebQuiz/credits.html">
           <span class="webquizlogo">WebQuiz</span><span class="TEX">T<span class="E">E</span>X</span>
        </a>
        <br>&copy; Copyright<br><span style="overflow: visible;">{copyright_years}</span>
      </div>
    </div>'''

question_buttons = r'''
      <div class="buttons">
        <br>{buttons}
      </div>
      <table class="marking-key">
         <tr><td class="star">&starf;</td><td style="width: 14ex;">{side_menu_star}</td></tr>
         <tr><td class="tick">&check;</td><td>{side_menu_tick}</td></tr>
         <tr><td class="cross">&cross;</td><td>{side_menu_cross}</td></tr>
      </table>'''

# quiz title and navigation arrows
quiz_header = r'''<div class="quiz-header">
       <div class="quiz-title">{title}</div><div></div>{arrows}
      </div>'''
navigation_arrows = r'''
       <span id="question-label" class="question-label">{question}</span>
       <span id="question-number" class="question-label">{question_number}</span>
       <span class="arrows">
          <a onClick="nextQuestion(-1);" title="{previous_question}">&#x25c4;</a>
          <span class="question-label">{questions}</span>
          <a onClick="nextQuestion(1);"  title="{next_question}">&#x25ba;</a>
       </span>'''

# discussion item
discussion = r'''<div id="question-{dnum}" class="question" style="display:{display};">
        {heading}{discussion.text}
      </div>
'''
discussion_heading = r'''<div class="question-label">{}</div>
        '''

#quiz index
quiz_index_div = r'''     <div class="quiz-index">
        <ul>
          {quiz_index}
        </ul>
      </div>'''
index_item = r'''<li><a href={url}>{title}</a></li>'''

# now we come to the question wrappers
question_wrapper = r'''<div id="question{qnum}" class="question" style="display:{display};">
      <span class="question-label">{question_number}</span>{question}
      {feedback}
      </div>
'''

question_text = r'''  {question_text}
      <form id="Q{qnum}Form" onSubmit="return false;">
        {question_options}
        <p>
          <input type="button" value="{check_answer}" name="answer" class="input-button" onClick="checkAnswer({qnum});"/>
          {nextquestion}
        </p>
      </form>
'''
nextquestion='<input type="button" value="{next_question}" class="input-button" title="{next_question}" name="next" onClick="nextQuestion(1);"/>'

# Questions and feedback:
input_answer = '{answer}&nbsp;<input type="text"  onChange="checkAnswer({qnum});" size="{size}"/>{after_text}'
choice_answer = '<table class="question-choices">{choices}</table>{after_text}'
input_single = '\n<input type="hidden" name="Q{qnum}hidden"/>'

single_item = '''<td><input type="radio" name="Q{qnum}option"/></td>
<td class="brown" >{choice}</td><td><div class="question-choices">{text}</div></td>
'''
multiple_item = '''<td><input type="checkbox" name="Q{qnum}option{optnum}"/></td>
<td class="brown" >{choice}</td><td><div class="question-choices">{text}</div></td>
'''

tf_feedback_text = r'''
        <div id="q{choice}{feedback}" class="feedback"><em class="dazzle">{correct_answer}</em> <em>{answer2}</em>
           <div>{text}</div>
        </div>'''
single_feedback = r'''
        <div id="q{qnum}feedback{part}" class="feedback">
              <em>{alpha_choice} <span class="dazzle">{correct_answer}</span></em>
              <div>{feedback}</div>
        </div>'''

multiple_feedback = r'''
        <div id="q{qnum}feedback{part}" class="feedback">
            <em>{one_mistake}</em><br>{multiple_choice_opener} <span class="dazzle">{correct_answer}</span>.
            <div>{feedback}</div>
        </div>'''
multiple_feedback_correct = r'''
        <div id="q{qnum}feedback0" class="feedback"><em class="dazzle">{correct}</em>
            <ol>
{feedback}
            </ol>
        </div>'''
multiple_feedback_answer = '              <li><em>{correct_answer}</em> {reason}</li>'

# no script error when javascript is not enabled
no_script = r'''<noscript>
    <div class="warning">
      If you are reading this message either your browser does not support
      JavaScript or because JavaScript is not enabled. You will need to enable
      JavaScript and then reload this page in order to use this quiz.
     </div>
  </noscript>'''

######################################################################
# missing webquiz.ini file
######################################################################

missing_webquiz_ini = r'''
webquiz installation error: unable to find webquiz.ini -> {}

If you believe that WebQuiz has been installed correctly then please try
rebuilding the TeX filename database. On TeXLive systems this can be done
using the command mktexlsl, or using the TeXLive gui. On MiKTEX you can
rebuild the database from the Tasks menu on the MiKTeX console. You may
need to run this command using an administrator account or sudo, on unix
systems.

'''

######################################################################
# the remaining templates prompt the user when initialising webquiz
######################################################################

initialise_invite = r'''WebQuiz needs to be initialised.

Before WebQuiz can display quiz web pages on your system it first needs to
copy some css and javascript files into a directory that is accessible from
your webserver.

Do you want to initialise WebQuiz now [Y/n]? '''

initialise_before_settings = r'''
Before you can view the webquiz settings you need to initialise WebQuiz
using the command:
    webquiz --initialise
'''

initialise_introduction = r'''
WebQuiz Initialisation
======================
If you want to install the web components of WebQuiz files into a "system"
directory then you should quit this program (use control-C on unix-like
systems) and then run
    webquiz --initialise
from an administrators account or, on unix-like systems, use:
    sudo webquiz --initialise

If you want to continue then WebQuiz will guide you through the
initialisation process.  For more information about the steps involved
see Section 3.2 of the WebQuiz manual On many systems, you can open
the webquiz manual using the command: texdoc webquiz
'''

webroot_request = r'''----
To make files accessible from your web server WebQuiz needs:
  o A directory, or folder, on your server that is visible from your
    web server. This directory MUST be accessible from the web. It can
    either be a "system" directory or in your personal web directories.
  o The relative URL for accessing these files from the web. This is
    the part of the URL that you have to add to your "root" URL to
    access the files. For example, if the URL for your department is
        http://www.maths.usyd.edu.au/
    and the WebQuiz files can be accessed as
        http://www.maths.usyd.edu.au/teaching/WebQuiz
    then the relative URL for the WebQuiz files is /teaching/WebQuiz

It is recommended that you create a separate directory for WebQuiz on your
web server. The location of the files on your web server will depend on
your operating system and system configuration. A common sysem location
for the WebQuiz web directory on a {platform} system is
     {webquiz_dir}
'''

not_installed = r'''
According to kpsewhich, the TeX components of WebQuiz are not installed
on your system. If you have downloaded the WebQuiz zip file from ctan
then try using:
    > webquiz --tex-install
If the TeX components are already installed then something has gone
horribly wrong. If you think this is a bug please report it by creating
an issue at {}
'''

oserror_copying = r'''There was a problem copying files to {web_dir}
Error: {err}

Please give a different directory
'''

webquiz_url_message = r'''
Please give the relative URL for the WebQuiz web directory. For example,
if the base of your web server's directory is /var/www/html/ and the
WebQuiz web directory is /var/www/html/courses/WebQuiz then the
relative URL for the WebQuiz directory would be /courses/WebQuiz.

WebQuiz relative URL [{}]: '''

initialise_ending = r'''
You should now be able to build web pages using webquiz! As an initial
test you can try to build the example files from the webquiz manual
by going to the directory
    {web_dir}/doc/examples
If pstricks and dvisgvm are properly configured (see Section 3.3 of
the manual), then you can also try building the online manual, by
going to the directory
    {web_dir}/doc/examples
and typing:
    webquiz webquiz-online-manual

If you want to change the default webquiz settings please use:
    webquiz --edit-settings
'''

text_initialise_warning = r'''
WebQuiz has not yet been initialised. To remove this warning please use
      webquiz --initialise
to install the WebQuiz web files onto your system.
'''

web_initialise_warning = r'''
      <div id="initialisewarning" class="warning">
        WebQuiz has not yet been initialised. To remove this warning box please use
        <blockquote>
          webquiz --initialise
        </blockquote>
        to install the WebQuiz web files on your system.
        <button style="float: right" onclick="document.getElementById('initialisewarning').style.display='none'">OK</button>
      </div>
'''

rc_permission_error = r'''
Attempting to write the rcfile to {rcfile}
resulted in the error:
  {error}

To write the WebQuiz rcfile into this directory you may need to quit
and run webquiz again, either using an administrator account, or using
sudo on linux/macosx.

Press the key 1-3, followed by RETURN, to:
    1. Try to save to {rcfile} again
    2. Save to the user rcfile {alt_rcfile}
    3. Give a different location for the rcfile
Press any other key to exit without saving.
'''

permission_error = r'''
You do not have permission to write to {}.

To install WebQuiz files into this directory you may need to quit and
run webquiz using either an administrator account, or using sudo on
linux/macosx.

Alternatively, please give a different directory.
'''

insufficient_permissions = r'''
Insufficient permissions. Try using sudo or using an admisitrator account.
{}
'''

webquiz_url_warning = r'''
WARNING: most of the time, but not always, the relative URL will be a
suffix of the web directory name, which is not the case with your
settings. Your URL may well be correct, however, if you have made a
mistake then you can change this at any time using the command:
    webquiz --initialise
'''

edit_settings = r'''----
In the webquizrc file you can set global defaults for the following:
    breadrumbs
    colour theme
    department
    department_url
    engine
    institution
    institution
    language
Leave these blank if you do not want default values. There are three
additional "advanced: options:
    mathjax
    webquiz_layout
    webquiz_mk4
Please only change these settings if you know what you are doing
because incorrect values for these settings will break WebQuiz.

With each setting the default value is printed inside square brackets.
For example, as [default]. Press RETURN to accept the default value.
Except for the location of the WebQuiz web directory all of these
settings are easy to change at any time using the command:
    webquiz --edit-settings
'''

advanced_settings = r'''
** The remaining settings are advanced settings that you probably **
** do not want to change at this time. Incorrect values for these **
** settings will break WebQuiz so accept the defaults unless you  **
** know what you are doing.                                       **
'''

# and now for the webquiz help message
webquiz_help_message = r'''
WebQuiz allows you to write interactive online quizzes using latex
'''
